"""

"""

from epyk.core.html.entities import EntHtml4
from epyk.core.html.entities import EntHtml5_A
from epyk.core.html.entities import EntHtml5_B
from epyk.core.html.entities import EntHtml5_C
from epyk.core.html.entities import EntHtml5_D

from epyk.core.html import Defaults


class Entities(object):
  """

  """

  non_breaking_space = '&#160;'
  less_than = '&#60;'
  greater_than = '&#62;'
  ampersand = '&#38;'
  double_quotation_mark = '&#34;'
  single_quotation_mark_apostrophe = '&#39;'
  cent = '&#162;'
  pound = '&#163;'
  yen = '&#165;'
  euro = '&#8364;'
  copyright = '&#169;'
  registered_trademark = '&#174;'

  @property
  def html4(self):
    """

    Related Pages:

			https://www.w3schools.com/charsets/ref_html_entities_4.asp
    """
    return EntHtml4

  @property
  def html5_a(self):
    """

    Related Pages:

			https://www.w3schools.com/charsets/ref_html_entities_a.asp
    """
    return EntHtml5_A

  @property
  def html5_b(self):
    """

    Related Pages:

			https://www.w3schools.com/charsets/ref_html_entities_b.asp
    """
    return EntHtml5_B

  @property
  def html5_c(self):
    """

    Related Pages:

			https://www.w3schools.com/charsets/ref_html_entities_c.asp
    """
    return EntHtml5_C

  @property
  def html5_d(self):
    """

    Related Pages:

			https://www.w3schools.com/charsets/ref_html_entities_d.asp
    """
    return EntHtml5_D

  @property
  def exts(self):
    """
    Bespoke Entities extensions from the Default layer
    """
    return Defaults.ENTITIES_ADD_ON



"""


"""


class Theme(object):

  def __init__(self, ovrs_attrs=None):
    self.__colors = {
      "charts": list(self._charts),
      "colors": list(self._colors),
      "greys": list(self._greys),
      "warning": list(self._warning),
      "danger": list(self._danger),
      "success": list(self._success),
    }
    if ovrs_attrs is not None:
      self.__colors.update(ovrs_attrs)

  @property
  def charts(self):
    return self.__colors["charts"]

  @charts.setter
  def charts(self, colors):
    self.__colors["charts"] = colors

  @property
  def colors(self):
    return self.__colors["colors"]

  @colors.setter
  def colors(self, colors):
    self.__colors["colors"] = colors

  @property
  def greys(self):
    return self.__colors["greys"]

  @greys.setter
  def greys(self, colors):
    self.__colors["greys"] = colors

  @property
  def warning(self):
    return self.__colors["warning"]

  @warning.setter
  def warning(self, colors):
    self.__colors["warning"] = colors

  @property
  def danger(self):
    return self.__colors["danger"]

  @danger.setter
  def danger(self, colors):
    self.__colors["danger"] = colors

  @property
  def success(self):
    return self.__colors["success"]

  @success.setter
  def success(self, colors):
    self.__colors["success"] = colors


class ThemeCustome(Theme):
  _charts, _colors, _greys = [], [], []
  _warning, _danger, _success = set(), set(), set()


class ThemeDefault(Theme):

  _charts = [
    '#009999', '#336699', '#ffdcb9',
    '#cc99ff', '#b3d9ff', '#ffff99',
    '#000066', '#b2dfdb', '#80cbc4',
    '#e0f2f1', '#b2dfdb', '#80cbc4',  # teal
    '#ffebee', '#ffcdd2', '#ef9a9a',  # red
    '#f3e5f5', '#e1bee7', '#ce93d8',  # purple
    '#ede7f6', '#d1c4e9', '#b39ddb',  # deep purple
    '#e8eaf6', '#c5cae9', '#9fa8da',  # indigo
    '#fffde7', '#fff9c4', '#fff59d',  # yellow
    '#fff3e0', '#ffe0b2', '#ffcc80',  # orange
    '#efebe9', '#d7ccc8', '#bcaaa4',  # brown
  ]
  _colors = ["#e8f5e9", '#c8e6c9', '#a5d6a7', '#81c784', '#66bb6a', '#4caf50', '#43a047', '#388e3c', '#2e7d32',
             '#1b5e20']
  _greys = ['#FFFFFF', '#f5f5f5', '#eeeeee', '#e0e0e0', '#bdbdbd', '#9e9e9e', '#757575', '#616161', '#424242',
            '#212121', '#000000']
  _warning, _danger, _success = ('#FFF3CD', '#e2ac00'), ("#F8D7DA", "#C00000"), ('#e8f2ef', '#3bb194')




from epyk.core.js.configs import JsConfig
from epyk.core import Page


report = Page.Report()

chart_type = 'hbar'
configs = JsConfig.get(chart_type, chartFam='Billboard', preferred=False)
config_chart = configs['Billboard'][chart_type](report, [], {})
config_chart.set_attrs(name="test", value='function(){alert}')

print(config_chart.toJs())

# Billboard
#
# lineChart = JsLine(None, [], {})
# lineChart.addAttr('pattern', ['yellow'], 'color')
# print(lineChart)
#
#
#
# # C3
# lineChart = JsLine(None, [], {})
# lineChart.addAttr('pattern', ['yellow'], 'color')
# print(lineChart)


# Check if pandas is available in the current environment
# if it is the case this module can handle Dataframe directly
try:
  import pandas as pd
  has_pandas = True

except:
  has_pandas = False

from epyk.core import html


class Lists(object):

  def __init__(self, context):
    self.context = context

  def _filter(self, recordSet, column, options=None):
    """
    Description:
    ------------



    Attributes:
    ----------
    :param recordSet:
    :param column:
    :param options: A dictionary with specific filtering options e.g {'allSelected': True, 'operation': 'in'}
    """
    dataId = id(recordSet)
    dataCode = "df_code_%s" % dataId
    globalFilter = {'jsId': dataCode, 'colName': column}
    globalFilter.update({options})
    if not dataCode in self.context.rptObj.jsSources:
      self.context.rptObj.jsSources[dataCode] = {'dataId': dataId, 'containers': [], 'data': recordSet}
      self.context.rptObj.jsSources[dataCode]['containers'].append(self)
    return globalFilter

  def _recordSet(self, recordSet, column):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param recordSet:
    :param column:
    """
    data = None
    is_converted = False
    if has_pandas:
      if isinstance(recordSet, pd.DataFrame):
        data = [{'name': r, 'value': r} for r in recordSet[column].unique().tolist()]
        is_converted = True

    if not is_converted:
      result = {}
      for rec in recordSet:
        result[rec[column]] = {'name': rec[column], 'value': rec[column]}
      data = [result[k] for k in sorted(result.keys())]
    return data

  def select(self, records=None, htmlCode=None, label=None, selected=None, width=(100, "%"), height=(None, "%"),
             column=None, filter=None, profile=None, multiple=False, options=None):
    """
    Description:
    ------------
    HTML Select component

    Usage::

      rptObj.ui.select(["A", "B", "C"], label="label", selected="C", multiple=True,
                      options={"title": "ttle", 'showTick': True, 'maxOptions': 2})
      s.selected = "B"
      s.change(rptObj.js.console.log(s.dom.val))

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlSelect.Select`

    Related Pages:

			https://silviomoreto.github.io/bootstrap-select/examples/
    https://www.npmjs.com/package/bootstrap-select-v4
    https://www.jqueryscript.net/form/Bootstrap-4-Dropdown-Select-Plugin-jQuery.html

    Attributes:
    ----------
    :param records: The input data. Can be a list or a dataFrame
    :param htmlCode: Optional. The component identifier code (for bot
    :param label: Optional. The HTML label attached to the component
    :param selected: The selected values
    :param width: Optional. Integer for the component width
    :param height: Optional. Integer for the component height
    :param column:
    :param filter:
    :param profile: Optional. A flag to set the component performance storage
    :param multiple: Boolean. To set if the component can handle multiple selections
    :param options: The select options as defined https://developer.snapappointments.com/bootstrap-select/options/
    """
    records = records or []
    options = {} if options is None else options

    all_selected = options.get("allSelected", False)
    empty_selected = options.get("empty_selected", True)
    if column is not None:
      if filter is not None:
        if filter == True:
          self._filter(records, column, {'allSelected': all_selected, 'operation': options.get("operation", "in")})
      records = self._recordSet(records, column)
    elif isinstance(records, (list, tuple)) and len(records) > 0 and not isinstance(records[0], dict):
      records = [{'name': rec, 'value': rec} for rec in records]
    if all_selected:
      records = [{'name': 'All', 'value': ''}] + records
    if empty_selected:
      records = [{'name': '', 'value': ''}] + records
    if multiple:
      if not isinstance(multiple, dict):
        multiple = {"max": 2}
      if selected is not None:
        for rec in records:
          if rec["value"] in selected:
            rec["selected"] = True
      return self.context.register(
        html.HtmlSelect.Select(self.context.rptObj, records, htmlCode, width, height, filter, profile, multiple, options))

    if selected is not None:
      for rec in records:
        if rec["value"] == selected:
          rec["selected"] = True
    html_select = html.HtmlSelect.Select(self.context.rptObj, records, htmlCode, width, height, filter, profile, multiple, options)
    self.context.register(html_select)
    return html_select

  def lookup(self, lookupData=None, htmlCode=None, label=None, selected=None, width=(100, "%"), height=(None, "%"),
             column=None, filter=None, profile=None, multiple=False, options=None):
    """
    Description:
    ------------
    HTML Select component

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlSelect.Lookup`

    Related Pages:

			https://silviomoreto.github.io/bootstrap-select/examples/
    https://www.npmjs.com/package/bootstrap-select-v4
    https://www.jqueryscript.net/form/Bootstrap-4-Dropdown-Select-Plugin-jQuery.html

    Attributes:
    ----------
    :param records: The input data. Can be a list or a dataFrame
    :param htmlCode: Optional. The component identifier code (for bot
    :param label: Optional. The HTML label attached to the component
    :param selected: The selected values
    :param width: Optional. Integer for the component width
    :param height: Optional. Integer for the component height
    :param column:
    :param filter:
    :param profile: Optional. A flag to set the component performance storage
    :param multiple: Boolean. To set if the component can handle multiple selections
    :param options: The select options as defined https://developer.snapappointments.com/bootstrap-select/options/
    """
    options = {} if options is None else options
    html_select = html.HtmlSelect.Lookup(self.context.rptObj, lookupData, htmlCode, width, height, filter, profile, multiple, options)
    self.context.register(html_select)
    return html_select

  def item(self, text=None):
    """
    Description:
    ------------

    Usage::

      l = rptObj.ui.lists.list(["A", "B"])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.List`

    Related Pages:

      https://www.w3schools.com/bootstrap/bootstrap_list_groups.asp
    http://astronautweb.co/snippet/font-awesome/
    """
    html_item = html.HtmlList.Li(self.context.rptObj, text)
    self.context.register(html_item)
    return html_item

  def list(self, data=None, color=None, width=('auto', ""), height=(None, 'px'),
           htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      l = rptObj.ui.lists.list(["A", "B"])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.List`

    Related Pages:

			https://www.w3schools.com/bootstrap/bootstrap_list_groups.asp
    http://astronautweb.co/snippet/font-awesome/
    """
    html_list = html.HtmlList.List(self.context.rptObj, data or [], color, width, height, htmlCode,
                                   helper, options or {}, profile)
    self.context.register(html_list)
    html_list.css({"list-style": 'none'})
    return html_list

  def numbers(self, data=None, color=None, width=('auto', ""), height=(None, 'px'),
                  htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.lists.numbers(["A", "B"])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.List`

    Related Pages:

        https://www.w3schools.com/html/html_lists.asp
    https://www.w3.org/wiki/CSS/Properties/list-style-type
    """
    html_list = html.HtmlList.List(self.context.rptObj, data or [], color, width, height, htmlCode,
                                   helper, options or {}, profile)
    self.context.register(html_list)
    html_list.css({"list-style-type": 'decimal'})
    return html_list

  def alpha(self, data=None, color=None, width=('auto', ""), height=(None, 'px'),
                  htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.List`

    :param data:
    :param color:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param options:
    :param profile:
    """
    html_list = html.HtmlList.List(self.context.rptObj, data or [], color, width, height, htmlCode,
                                   helper, options or {}, profile)
    self.context.register(html_list)
    html_list.css({"list-style-type": 'lower-alpha'})
    return html_list

  def roman(self, data=None, color=None, width=('auto', ""), height=(None, 'px'),
                  htmlCode=None, helper=None, options=None, profile=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.List`

    :param data:
    :param color:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param options:
    :param profile:
    :return:
    """

    html_list = html.HtmlList.List(self.context.rptObj, data or [], color, width, height, htmlCode,
                                   helper, options or {}, profile)
    self.context.register(html_list)
    html_list.css({"list-style-type": 'lower-roman'})
    return html_list

  def points(self, data=None, color=None, width=('auto', ""), height=(None, 'px'),
             htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.List`

    Related Pages:

			https://www.w3schools.com/html/html_lists.asp

    Attributes:
    ----------
    :param data:
    :param color:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param options:
    :param profile:
    """
    html_list = html.HtmlList.List(self.context.rptObj, data or [], color, width, height, htmlCode,
                                   helper, options or {}, profile)
    self.context.register(html_list)
    html_list.css({"list-style-type": 'circle'})
    return html_list

  def disc(self, data=None, color=None, width=('auto', ""), height=(None, 'px'), htmlCode=None, helper=None,
           options=None, profile=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.List`

    Related Pages:

			https://www.w3schools.com/cssref/pr_list-style-type.asp

    Attributes:
    ----------
    :param data:
    :param color:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param options:
    :param profile:
    """
    html_list = html.HtmlList.List(self.context.rptObj, data or [], color, width, height, htmlCode,
                                   helper, options or {}, profile)
    self.context.register(html_list)
    html_list.css({"list-style-type": 'disc'})
    return html_list

  def squares(self, data=None, color=None, width=('auto', ""), height=(None, 'px'),
             htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.lists.squares(["A", "B"])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.List`

    Related Pages:

        https://www.w3schools.com/cssref/pr_list-style-type.asp

    Attributes:
    ----------
    :param data:
    :param color:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param options:
    :param profile:
    """
    html_list = html.HtmlList.List(self.context.rptObj, data or [], color, width, height, htmlCode,
                                   helper, options or {}, profile)
    self.context.register(html_list)
    html_list.css({"list-style-type": 'square'})
    return html_list

  def groups(self, data=None, categories=None, color=None, width=('auto', ""), height=(None, 'px'),
             htmlCode=None, helper=None, profile=None):
    """
    Description:
    ------------

    Usage::

      l = rptObj.ui.lists.groups(["AWW", "B"])
      l.add_list(["D", "E"], category="Test")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.Groups`

    Related Pages:

    http://designbump.com/create-a-vertical-accordion-menu-using-css3-tutorial/
    http://thecodeplayer.com/walkthrough/vertical-accordion-menu-using-jquery-css3

    Attributes:
    ----------
    :param data:
    :param categories:
    :param color:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param profile:
    """
    data = data or []
    categories = categories or [""]
    if len(data) > 0:
      if isinstance(data[0], list):
        categories = [""] * len(data)
      else:
        # This object is expecting a list of lists
        data = [data]
    html_obj = html.HtmlList.Groups(self.context.rptObj, data, categories, color, width, height, htmlCode,
                                    helper, profile)
    self.context.register(html_obj)
    return html_obj

  def checklist(self, data=None, color=None, width=('auto', ""), height=(None, 'px'),
                htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      data = [{"label": "python", "value": False}, {"label": "Java", "value": 5}]
      checks = rptObj.ui.lists.checklist(data)

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.Checks`

    Related Pages:

    ----------
    :param data:
    :param color:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param profile:
    """
    html_obj = html.HtmlList.Checks(self.context.rptObj, data or [], color, width, height, htmlCode, helper, options or {}, profile)
    self.context.register(html_obj)
    return html_obj

  def tree(self, data=None, color=None, width=('auto', ""), height=(None, 'px'),
           htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      data = [{"label": 'test', 'items': [{"label": 'child 1', 'color': 'red'}]}]
      rptObj.ui.lists.tree(data)

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTrees.Tree`

    ----------
    :param data:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    html_tree = html.HtmlTrees.Tree(self.context.rptObj, data or [], color, width, height, htmlCode, helper, options, profile)
    self.context.register(html_tree)
    return html_tree

  def dropdown(self, recordSet=None, text="", width=('auto', ""), height=(32, 'px'), htmlCode=None,
               helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlTrees.DropDown`

    Related Pages:

      http://getbootstrap.com/docs/4.0/components/dropdowns/
      https://www.w3schools.com/bootstrap/tryit.asp?filename=trybs_ref_js_dropdown_multilevel_css&stacked=h
      https://codepen.io/svnt/pen/beEgre

    Attributes:
    ----------
    :param recordSet:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    dftl_options = {"width": 70}
    dftl_options.update(options or {})
    html_d = html.HtmlTrees.DropDown(self.context.rptObj, recordSet, text, width, height, htmlCode, helper,
                                     dftl_options, profile)
    self.context.register(html_d)
    return html_d

  def badges(self, data=None, color=None, width=('auto', ""), height=(None, 'px'),
             htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.lists.badges([{'label': 'Python', 'value': 12}, {'label': 'R', 'value': 3}])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.Badges`

    Related Pages:

      https://www.w3schools.com/bootstrap/bootstrap_list_groups.asp
      https://v4-alpha.getbootstrap.com/components/list-group/

    Attributes:
    ----------
    :param data:
    :param color:
    :param width:
    :param height:
    :param options:
    :param profile:
    """
    html_obj = html.HtmlList.Badges(self.context.rptObj, data or [], color, width, height, htmlCode, helper, options or {}, profile)
    self.context.register(html_obj)
    return html_obj

  def buttons(self, data=None, color=None, width=('auto', ""), height=(None, 'px'),
              htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.lists.badges([{'label': 'Python', 'value': 12}, {'label': 'R', 'value': 3}])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.Buttons`

    Related Pages:

      https://www.w3schools.com/bootstrap/bootstrap_list_groups.asp
      https://v4-alpha.getbootstrap.com/components/list-group/

    Attributes:
    ----------
    :param data:
    :param color:
    :param width:
    :param height:
    :param profile:
    """
    html_obj = html.HtmlList.Buttons(self.context.rptObj, data or [], color, width, height, htmlCode, helper, options or {}, profile)

    self.context.register(html_obj)
    return html_obj

  def radios(self, records, group_name=None, width=('auto', ""), height=(None, "px"), htmlCode=None, helper=None,
             options=None, profile=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlList.ListTournaments`
      - :class:`epyk.core.html.HtmlContainer.Div`
      - :class:`epyk.core.html.HtmlInput.Radio`



    :param records:
    :param group_name:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param options:
    :param profile:
    """
    container = self.context.rptObj.ui.div(width=width, height=height, htmlCode=htmlCode, helper=helper, profile=profile)
    group_name = group_name or container.htmlId
    for rec in records:
      container += self.context.rptObj.ui.fields.radio(rec.get("value", False), rec.get('label', ''), rec.get("group_name", group_name))
    return container

  def brackets(self, recordSet=None, width=('auto', ""), height=(550, 'px'), options=None, profile=None):
    return self.context.register(html.HtmlList.ListTournaments(self.context.rptObj, recordSet, width, height, options, profile))

  def tags(self, items=None, title="Filter Panel", width=(100, "%"), height=(None, "px"), htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlEvent.Filters`

    Attributes:
    ----------
    :param items:
    :param title:
    :param width:
    :param height:
    :param htmlCode:
    :param helper:
    :param profile:
    """
    html_f = html.HtmlEvent.Filters(self.context.rptObj, items or [], title, width, height, htmlCode, helper, options or {}, profile)
    self.context.register(html_f)
    return html_f



"""
Module dedicated to wrap the Javascript Object

"""

import json

from epyk.core.js import JsUtils

_JSVARS = 0


class JsKeyword(object):
  def __init__(self, keyword):
    self.__keyword = keyword

  def toStr(self):
    return self.__keyword

  def __str__(self):
    return self.__keyword


class JsObject(object):
  _jsClass = "Object"

  def __init__(self, data, varName=None, setVar=False, isPyData=False, report=None):
    """

    Related Pages:

			https//www.w3schools.com/js/js_type_conversion.asp

    Attributes:
    ----------
    :param varName:
    :param data:
    :param setVar:
    :param isPyData:
    :param report: The internal report object

    """
    global _JSVARS

    self.varName, self.varData, self._js, self._report = varName, str(data), [], report
    self._frozen, self._sealed = False, False
    if varName is None and setVar:
      _JSVARS += 1
      self.varName = "%s_%s" % (self.__class__.__name__, _JSVARS)
    if setVar:
      self.setVar(self.varName)

  @classmethod
  def new(cls, data=None, varName=None, isPyData=True, report=None):
    """
    Create a Python Javascript Object

    Example
    JsDate.new("2019-05-03", varName="MyDate")

    Related Pages:

			https://www.w3schools.com/jsref/jsref_obj_date.asp

    :param data: Optional, The object data
    :param varName: Optional, The object variable name
    :param isPyData: Optional, Boolean to specify if it is a Python reference and if it should be converted to Json
    :param report: The internal report object

    :return: The Python Javascript Date primitive
    """
    if isPyData:
      return cls(data=JsUtils.jsConvertData(data, None), varName=varName, setVar=True, isPyData=isPyData, report=report)

    return cls(data=data, varName=varName, setVar=True, isPyData=isPyData, report=report)

  @classmethod
  def this(cls, report=None):
    """
    Get the object this

    Example
    jsObj.object.this

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/this

    :param report: The internal report object

    :return: The python Javascript object
    """
    return cls.get("this", report=report)

  @classmethod
  def get(cls, varName, report=None):
    """
    Get the Javascript Object by its reference

    Example
    JsDate.new("2019-05-03", varName="MyDate")
    JsDate.get("MyDate")

    Related Pages:

			https://www.w3schools.com/jsref/jsref_obj_date.asp

    :param varName: The Javascript object reference
    :param report: The internal report object

    :return: The python Javascript object
    """
    return cls(data=None, varName=varName, setVar=False, report=report)

  @property
  def varId(self):
    """
    The Javascript and Python reference ID

    :return: The Javascript String of the object variable name
    """
    return self.varData if self.varName is None else self.varName

  def setVar(self, varName, varType="var"):
    """
    The setVar() method will defined the variable name and use this reference in the future

    Example

    Documentation

    :param varName: Required, The variable name
    :param varType: The type of variable to be set on the Javascript side

    :return: The Python Javascript Object
    """
    if varName is None:
      return self

    if varName != self.varName:
      if len(self._js) == 1:
        if self._js[0].startswith("var "):
          # Remove the entry used to define by default the javascript object
          self._js = ["%s %s = %s" % (varType, varName, self.varData)]
        else:
          self._js.append("%s %s = %s" % (varType, varName, self.varName))
      else:
        self._js.append("%s %s = %s" % (varType, varName, self.varName or self.varData))
    else:
      self._js.append("%s %s = %s" % (varType, varName, self.varData))
    return self

  def prototype(self, name, value):
    """
    The prototype property allows you to add new properties and methods to existing object types.

    Documentation:
      - https://www.w3schools.com/jsref/jsref_prototype_string.asp

    :param name: The objects property name
    :param value: The objects property values

    :return: A reference to the String.prototype object
    """
    return "%s.prototype.%s = %s" % (self.varName, name, value)

  def add(self, n):
    """
    Add value to a Javascript Number.
    The value will be added and it will return a new number object on the Javascript side

    Example
    jsNumber.add(34.5)

    Documentation

    :param n: The number value

    :return: A new Python Javascript Number
    """
    jsData = JsUtils.jsConvertData(n, None)
    return JsObject("%s + %s" % (self.varId, jsData), isPyData=False)

  def __add__(self, value):
    return JsObject("%s += %s" % (self.varId, value), isPyData=False)

  def __sub__(self, value):
    return JsObject("%s -= %s" % (self.varId, value), isPyData=False)

  def __mul__(self, value):
    return JsObject("%s *= %s" % (self.varId, value), isPyData=False)

  def __truediv__(self, value):
    return JsObject("%s /= %s" % (self.varId, value), isPyData=False)

  def __mod__(self, value):
    return JsObject("%s %%= %s" % (self.varId, value), isPyData=False)

  def __pow__(self, value):
    from epyk.core.js import JsMaths

    return JsObject("%s = %s" % (self.varId, JsMaths.JsMaths.pow(self, value)))

  def __eq__(self, a):
    """

    :param a:

    :return:
    """
    if hasattr(a, 'toStr'):
      return "%s == %s" % (self.varId, a.toStr())

    if hasattr(a, 'varName'):
      return "%s == %s" % (self.varId, a.varName)

    return "%s == %s" % (self.varId, json.dumps(a))

  def __lt__(self, a):
    """

    :param a:
    :return:
    """
    if hasattr(a, 'toStr'):
      return "%s < %s" % (self.varId, a.toStr())

    if hasattr(a, 'varName'):
      return "%s < %s" % (self.varId, a.varName)

    return "%s < %s" % (self.varId, json.dumps(a))

  def __le__(self, a):
    """

    :param a:

    :return:
    """
    if hasattr(a, 'toStr'):
      return "%s <= %s" % (self.varId, a.toStr())

    if hasattr(a, 'varName'):
      return "%s <= %s" % (self.varId, a.varName)

    return "%s <= %s" % (self.varId, json.dumps(a))

  def __ne__(self, a):
    """

    :param a:

    :return:
    """
    if hasattr(a, 'toStr'):
      return "%s != %s" % (self.varId, a.toStr())

    if hasattr(a, 'varName'):
      return "%s != %s" % (self.varId, a.varName)

    return "%s != %s" % (self.varId, json.dumps(a))

  def __gt__(self, a):
    """

    :param a:

    :return:
    """
    if hasattr(a, 'toStr'):
      return "%s > %s" % (self.varId, a.toStr())

    if hasattr(a, 'varName'):
      return "%s > %s" % (self.varId, a.varName)

    return "%s > %s" % (self.varId, json.dumps(a))

  def __ge__(self, a):
    """

    :param a:

    :return:
    """
    if hasattr(a, 'toStr'):
      return "%s >= %s" % (self.varId, a.toStr())

    if hasattr(a, 'varName'):
      return "%s >= %s" % (self.varId, a.varName)

    return "%s >= %s" % (self.varId, json.dumps(a))

  def isFrozen(self):
    """
    The Object.isFrozen() determines if an object is frozen.

    Example
    jsObj.objects.get("MyObject").isFrozen()

    Related Pages:

			https://medium.com/@wlodarczyk_j/object-freeze-vs-object-seal-ba6d7553a436
    https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/isFrozen

    :return: A Python Javascript Boolean
    """
    if self.varName is None:
      raise Exception("Cannot freeze an object without variable name")

    from epyk.core.js.primitives import JsBoolean
    return JsBoolean.JsBoolean("Object.isFrozen(%s)" % self.varName, isPyData=False)

  def freeze(self):
    """
    The Object.freeze() method freezes an object. A frozen object can no longer be changed; freezing an object prevents new properties from being added to it, existing properties from being removed, prevents changing the enumerability, configurability, or writability of existing properties, and prevents the values of existing properties from being changed.

    Example
    jsObj.objects.get("MyObject").freeze()

    Related Pages:

			https://medium.com/@wlodarczyk_j/object-freeze-vs-object-seal-ba6d7553a436
    https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/freeze

    :return: The object that was passed to the function
    """
    if self.varName is None:
      raise Exception("Cannot freeze an object without variable name")

    self._frozen = True
    return JsObject("Object.freeze(%s)" % self.varName, isPyData=False)

  def isSealed(self):
    """
    The Object.seal() method seals an object, preventing new properties from being added to it and marking all existing properties as non-configurable.
    Values of present properties can still be changed as long as they are writable

    Example
    jsObj.objects.get("MyObject").isSealed()

    Related Pages:

			https://medium.com/@wlodarczyk_j/object-freeze-vs-object-seal-ba6d7553a436
    https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/seal

    :return: A Javascript Boolean
    """
    from epyk.core.js.primitives import JsBoolean

    self._sealed = True
    return JsBoolean.JsBoolean("Object.isSealed(%s)" % self.varId, isPyData=False)

  def defineProperty(self, obj, prop, descriptor):
    """
    The static method Object.defineProperty() defines a new property directly on an object, or modifies an existing property on an object, and returns the object.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/defineProperty

    :param obj: The object on which to define the property.
    :param prop: The name or Symbol of the property to be defined or modified.
    :param descriptor: The descriptor for the property being defined or modified.

    :return: The object that was passed to the function.
    """
    return JsObject("Object.defineProperty(%s, %s, %s)" % (obj.varId, JsUtils.jsConvertData(prop, None), JsUtils.jsConvertData(descriptor, None)))

  def getOwnPropertyNames(self, obj):
    """
    The Object.getOwnPropertyNames() method returns an array of all properties (including non-enumerable properties except for those which use Symbol) found directly in a given object.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/getOwnPropertyNames

    :param obj: The object whose enumerable and non-enumerable properties are to be returned.

    :return: An JsArray of strings that corresponds to the properties found directly in the given object.
    """
    from epyk.core.js.primitives import JsArray

    return JsArray.JsArray("Object.getOwnPropertyNames(%s)" % obj.varId)

  def seal(self):
    """
    The Object.seal() method seals an object, preventing new properties from being added to it and marking all existing properties as non-configurable.
    Values of present properties can still be changed as long as they are writable.

    Related Pages:

			https://medium.com/@wlodarczyk_j/object-freeze-vs-object-seal-ba6d7553a436
    https://www.w3schools.com/Js/js_object_es5.asp
    https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/seal

    :return: The Python Javascript object being sealed.
    """
    if self.varName is None:
      raise Exception("Variable name must be defined")

    return JsObject("Object.seal(%s)" % self.varName)

  def assign(self, target, sources, jsObj=None):
    """
    The Object.assign() method is used to copy the values of all enumerable own properties from one or more source objects to a target object.
    It will return the target object.
    This function might not work with older browser, so to guarantee a good compatibility the jsObj must be defined.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/assign

    :param target: The target object.
    :param sources: The source object(s).
    :param jsObj: Optional, The base Python Javascript object to add the polyfill

    :return:
    """
    if jsObj is not None:
      # Add a polyfill to ensure the browser compatibility
      jsObj._addImport("babel-polyfill")

    if not isinstance(sources, list):
      sources = [sources]
    srcObj = ",".join([JsUtils.jsConvertData(s, None) for s in sources])
    return JsObject("Object.assign(%s, %s)" % (JsUtils.jsConvertData(target, None), srcObj), isPyData=False)

  def create(self, proto=None, propertiesObject=None):
    """
    The Object.create() method creates a new object, using an existing object as the prototype of the newly created object.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/create

    :param proto: The object which should be the prototype of the newly-created object.
    :param propertiesObject: Optional. If specified and not undefined, an object whose enumerable own properties (that is, those properties defined upon itself and not enumerable properties along its prototype chain) specify property descriptors to be added to the newly-created object, with the corresponding property names.
                                       These properties correspond to the second argument of Object.defineProperties()

    :return: A Python Javascript object
    """
    if object is None:
      return JsObject("Object.create()")

    return JsObject("Object.create(%s)" % object)

  def entries(self):
    """
    The entries() method returns an Array Iterator object with key/value pairs.

    Related Pages:

			https://www.w3schools.com/jsref/jsref_entries.asp
    https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/entries

    :return: A new Array, containing the selected elements
    """
    from epyk.core.js.primitives import JsArray

    return JsArray.JsArray("Object.entries(%s)" % self.varId)

  def setattr(self, key, value):
    """

    Example
    jsObj.objects.get("MyObject").setattr("tesr", jsObj.objects.number.get("MyNumber"))

    Related Pages:

			https://www.w3schools.com/js/js_objects.asp
    https://www.w3schools.com/js/js_object_es5.asp

    :param key: The key to add to the object
    :param value: The value corresponding to the key. Can be a Python object or a Javascript reference

    :return: The Python Javascript object
    """
    if self.varName is None:
      raise Exception("Not name defined for this object")

    if self._frozen:
      print("Warning, try to change a frozen variable")

    return JsObject("%s[%s] = %s" % (self.varName, JsUtils.jsConvertData(key, None), JsUtils.jsConvertData(value, None)), setVar=False)

  def addItem(self, key, value):
    """
    Wrapper to the setattr method
    """
    return self.setattr(key, value)

  def __getitem__(self, key):
    """
    Return the value for a given key defined in the object

    Example
    jsObj.objects.get("MyObject")["test"]

    Related Pages:

			https://www.w3schools.com/js/js_object_es5.asp

    :param key: The String used as key

    :return: The corresponding Javascript object
    """
    return JsObject("%s[%s]" % (self.varId, JsUtils.jsConvertData(key, None)))

  def keys(self):
    """
    Returns an array of enumerable properties

    Example
    jsObj.objects.get("MyObject").keys()

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object/keys
    https://www.w3schools.com/js/js_object_es5.asp

    :return: A new Array, containing the selected elements
    """
    from epyk.core.js.primitives import JsArray

    return JsArray.JsArray("Object.keys(%s)" % self.varId, isPyData=False)

  def update(self, dico, jsObj=None):
    """
    This might not be supported by all the browser

    Documentation

    :param dico:
    :param jsObj: Optional, The base Python Javascript object to add the polyfill

    :return:
    """
    if jsObj is not None:
      # Add a polyfill to ensure the browser compatibility
      jsObj._addImport("babel-polyfill")
    return "try{Object.assign({}, %s, %s)} catch(err){console.warn('Assign not supported by the browser')}" % (self.varId, dico)

  def toString(self, explicit=True):
    """
    Converts a Object to a string, and returns the result

    Example
    jsObj.objects.get("MyObject").toString()

    Related Pages:

			https://www.w3schools.com/JS/js_number_methods.asp

    :param explicit: Optional, default True. Parameter to force the String conversion on the Js side

    :return: A Javascript String
    """
    from epyk.core.js.primitives import JsString

    if not explicit:
      return JsString.JsString("%s" % self.varId, isPyData=False)

    return JsString.JsString("%s.toString()" % self.varId, isPyData=False)

  def isArray(self):
    """
    The isArray() method determines whether an object is an array.

    Example
    jsObj.objects.get("MyObject").isArray()

    Related Pages:

			https://www.w3schools.com/jsref/jsref_isarray.asp

    :return: A Javascript Boolean
    """
    from epyk.core.js.primitives import JsBoolean

    return JsBoolean.JsBoolean("Array.isArray(%s)" % self.varId, isPyData=False, setVar=False)

  def toArray(self):
    """
    THis is not a standard Javascript method for an object.
    It is only defined for some objects like the Datatable data()

    :return:
    """
    from epyk.core.js.primitives import JsArray

    return JsArray.JsArray("%s.toArray()" % self.varId, isPyData=False, setVar=False)

  def toStr(self):
    """
    Internal function which return the object reference.
    This will either return the variable name if available and defined or the data itself

    :return: The Javascript String reference
    """
    result = list(self._js)
    self._js = []
    if len(result) > 0:
      return ";".join(JsUtils.jsConvertFncs(result))

    return self.varData if self.varName is None else self.varName

  def toRecord(self, header, varName):
    """

    Example
    d = rptObj.ui.div()
    d.drop([rptObj.js.objects.data.toRecord([1, 2, 3, 4], "result")])

    :param header:
    :param varName:

    :return:
    """
    from epyk.core.js.primitives import JsArray
    from epyk.core.js.objects import JsData
    from epyk.core.js.fncs.JsFncs import JsFunctions

    fncs = JsFunctions([
      self.toString().split("\\n").setVar("rows"), JsArray.JsArray.set(varName),
      JsArray.JsArray.get("rows").forEach([
        JsArray.JsArray.get(JsData.JsData(self._report).loop().val.toString().split("\t")).toDict(header).setVar("row").r,
        JsArray.JsArray.get(varName).push(JsObject.get("row"))])])
    record = JsObject.get(varName)
    record._js = [fncs.toStr()] + record._js
    return record

  @property
  def r(self):
    """
    Return the String representation of the Js object.
    This will produce the chain, empty the internal buffer and produce the string
    """
    return self.toStr()

  def __str__(self):
    """
    The str() method return the variable Javascript reference of the variable.

    According to the variable definition it can be either its name or its value

    :return: A Javascript reference
    """
    return self.varData if self.varName is None else self.varName




import re

from epyk.core.html import Html
from epyk.core.css import Colors

from epyk.core.html.options import OptText
from epyk.core.js.html import JsHtml

# The list of CSS classes
from epyk.core.css.styles import GrpCls
from epyk.core.css import Defaults_css

# The list of CSS classes
from epyk.core.css.styles import GrpClsText


class UpDown(Html.Html):
  name, category, callFnc = 'Up and Down', 'Texts', 'updown'
  __reqCss, __reqJs = ['font-awesome'], ['accounting', 'font-awesome']

  def __init__(self, report, rec, color, label, options, helper, profile):
    if rec is None:
      rec = {'value': 0, 'previous': 0}
    if label is not None:
      rec["label"] = label
    super(UpDown, self).__init__(report, rec, profile=profile)
    self.add_helper(helper)
    self.val['color'] = self._report.theme.colors[9] if color is None else color
    self.__options = OptText.OptionsNumber(self, options)

  @property
  def options(self):
    """
    Property to set all the possible object for a button

    :rtype: OptText.OptionsNumber
    """
    return self.__options

  @property
  def _js__builder__(self):
    return '''
      var delta = data.value - data.previous; 
      if(data.previous == 0) {var relMove = 'N/A'} else{var relMove = 100 * ((data.value - data.previous) / data.previous)};
      if(data.digits == undefined){data.digits = 0};
      if(data.label != undefined){htmlObj.append("<span style='padding:5px;font-size:"+ options.font_size+"'>"+ data.label +"</span>")};
      var valueElt = document.createElement('span'); valueElt.setAttribute('style', 'padding:5px'); 
      valueElt.innerHTML = accounting.formatNumber(data.value, options.digits, options.thousand_sep, options.decimal_sep); 
      htmlObj.appendChild(valueElt); var deltaElt = document.createElement('span');
      var relMoveElt = document.createElement('span'); var icon = document.createElement('i');
      if (delta < 0){
        deltaElt.setAttribute('style', 'padding:5px;color:'+ options.red +';font-size:'+ options.font_size);
        deltaElt.innerHTML = "(+"+ accounting.formatNumber(delta, options.digits, options.thousand_sep, options.decimal_sep) +")";
        relMoveElt.setAttribute('style', 'padding:5px;color:'+ options.red +';font-size:'+ options.font_size)
        relMoveElt.innerHTML = "("+ accounting.formatNumber(relMove, 2, options.thousand_sep, options.decimal_sep) +"%)";
        icon.className = 'fas fa-arrow-down';
        icon.setAttribute('style', 'color:'+ options.red +';font-size:'+ options.font_size)}
      else{  
        deltaElt.setAttribute('style', 'padding:5px;color:'+ options.green +';font-size:'+ options.font_size);
        deltaElt.innerHTML = "(+"+ accounting.formatNumber(delta, options.digits, options.thousand_sep, options.decimal_sep) +")";
        relMoveElt.setAttribute('style', 'padding:5px;color:'+ options.green +';font-size:'+ options.font_size)
        relMoveElt.innerHTML = "("+ accounting.formatNumber(relMove, 2, options.thousand_sep, options.decimal_sep) +"%)";
        icon.className = 'fas fa-arrow-up';
        icon.setAttribute('style', 'color:'+ options.green +';font-size:'+ options.font_size)};
      htmlObj.appendChild(deltaElt); htmlObj.appendChild(relMoveElt); htmlObj.appendChild(icon);
      '''

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<div %s></div>%s' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.helper)


class BlockText(Html.Html):
  __reqCss, __reqJs = ['font-awesome'], ['font-awesome']
  name, category, callFnc = 'Block text', 'Rich', 'blocktext'
  # _grpCls = CssGrpClsText.CssClassTextBlock

  def __init__(self, report, recordSet, color, border, width, height, helper, options, profile):
    super(BlockText, self).__init__(report, recordSet, css_attrs={'color': color, "width": width, "height": height}, profile=profile)
    self.add_helper(helper)
    self.__options = OptText.OptionsText(self, options)
    self.css({'padding': '5px'})
    if border != 'auto':
      self.css('border', str(border))

  @property
  def options(self):
    """
    Property to set all the possible object for a button

    :rtype: OptText.OptionsText
    """
    return self.__options

  @property
  def _js__builder__(self):
    return '''
      htmlObj.find('div').first().html(data.title); htmlObj.find('div').last().empty(); var content;
      if (typeof data.text === 'string' || data.text instanceof String) {content = data.text.split("\\n")}
      else {content = data.text}
      content.forEach(function(line){htmlObj.find('div').last().append('<p class="py_csstext">'+ line +'</a>')});
      if(options.showdown){var converter = new showdown.Converter(options.showdown); data.text = converter.makeHtml(data.text)} 
      htmlObj.find('div').last().html(data.text);
      if (data.color != undefined) {htmlObj.find('div').last().css('color', data.color)};
      if(typeof data.button != 'undefined'){
        htmlObj.find("a").html(data.button.text); htmlObj.find("a").attr('href', data.button.url)}
      '''

  def __str__(self):
    items = ['<div %s>' % self.get_attrs(pyClassNames=self.style.get_classes())]
    items.append('<div id="%s_title" style="font-size:%spx;text-align:left"><a></a></div>' % (self.htmlId, Defaults_css.font(3)))
    items.append('<div id="%s_p" style="width:100%%;text-justify:inter-word;text-align:justify;"></div>' % self.htmlId)
    #if self.val.get('button') is not None:
    #  items.append('<a href="#" %s><i></i></a>' % (self._report.style.getClsTag(['CssHrefNoDecoration', 'CssButtonBasic'])))
    items.append('</div>')
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return ''.join(items)


class TextWithBorder(Html.Html):
  __reqCss, __reqJs = ['font-awesome'], ['font-awesome']
  name, category, callFnc = 'Text with Border and Icon', 'Rich', 'textborder'

  def __init__(self, report, recordSet, width, height, align, helper, options, profile):
    super(TextWithBorder, self).__init__(report, recordSet, css_attrs={"width": width, "height": height}, profile=profile)
    self.add_helper(helper)
    self.__options = OptText.OptionsText(self, options)
    self.align = align
    if not 'colorTitle' in self.val:
      self.val['colorTitle'] = self._report.theme.colors[9]
    if not 'color' in self.val:
      self.val['color'] = self._report.theme.colors[9]
    self.css({"border-color": self.val['colorTitle'], 'margin-top': '20px'})

  @property
  def options(self):
    """
    Property to set all the possible object for a button

    :rtype: OptText.OptionsText
    """
    return self.__options

  @property
  def _js__builder__(self):
    return '''
      if(options.showdown){var converter = new showdown.Converter(options.showdown); 
        data.title = converter.makeHtml(data.title); data.value = converter.makeHtml(data.value)} 
      htmlObj.querySelector('legend').innerHTML = data.title; htmlObj.querySelector('span').innerHTML = data.value'''

  def __str__(self):
    item = ['<fieldset %s>' % self.get_attrs(pyClassNames=self.style.get_classes())]
    if 'icon' in self.val:
      self.val['align'] = self.align
      item.append('<i class="%(icon)s fa-5x" style="width:100%%;text-align:%(align)s;margin:2px 0 10px 0;color:%(color)s"></i>' % self.val)
    if 'url' in self.val:
      item.append('<legend style="font-size:%spx;color:%s"></legend><span></span><br><a style="float:right" href="%s">+ more details</a></fieldset>' % (Defaults_css.font(10), self.val['colorTitle'], self.val['url']))
    else:
      item.append('<legend style="font-size:%spx;color:%s"></legend><span></span></fieldset>' % (Defaults_css.font(10), self.val['colorTitle']))
    item.append(self.helper)
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return "".join(item)


class Number(Html.Html):
  name, category, callFnc = 'Number', 'Rich', 'number'

  def __init__(self, report, number, label, width, height, profile, options):
    super(Number, self).__init__(report, number, css_attrs={"width": width, "height": height}, profile=profile)
    if options.get('url', None) is not None:
      self.add_link(number, url=options['url'], css={"font-size": Defaults_css.font(10),
                      "width": "100%", 'text-decoration': 'none', 'display': 'inline-block', "text-align": 'center',
                      'margin': 0, 'color': 'inherit', 'padding': 0})
      self.span = self.link
    else:
      self.add_link(number, url="#", css={"font-size": Defaults_css.font(10),
                     "width": "100%", 'text-decoration': 'none', 'cursor': 'default',
                     'display': 'inline-block', "text-align": 'center', 'margin': 0, 'color': 'inherit', 'padding': 0})
      self.link.attr['target'] = '_self'
      self.span = self.link

      # self.add_span(number, css={'height': 'auto', "font-size": "%s%s" % (size[0] + 10, size[1]), "width": "100%", 'margin': 0})
    self.add_label(label, css={'float': 'none', "width": "100%", "margin": 0})
    self.css({"display": "inline-block", 'padding': '2px 0', 'clear': 'both', 'margin': '2px'})

  def __str__(self):
    return "<div %s></div>" % self.get_attrs(pyClassNames=self.style.get_classes())


class Delta(Html.Html):
  __reqCss, __reqJs = ['jqueryui'], ['accounting', 'jqueryui']
  name, category, callFnc = 'Delta Figures', 'Rich', 'delta'

  def __init__(self, report, records, width, height, options, helper, profile):
    super(Delta, self).__init__(report, records, css_attrs={"width": width, "height": height}, profile=profile)
    self.add_helper(helper)
    if not 'color' in self.val:
      self.val['color'] = self._report.theme.colors[9]
    if not 'thresold1' in self.val:
      self.val['thresold1'] = 100
    if not 'thresold2' in self.val:
      self.val['thresold2'] = 50
    self.css({"color": self.val['color']})
    self.__options = OptText.OptionsNumber(self, options)

  @property
  def options(self):
    """
    Property to set all the possible object for a button

    :rtype: OptText.OptionsNumber
    """
    return self.__options

  @property
  def _js__builder__(self):
    return '''
       jHtmlObj = jQuery(htmlObj);
       var variation = 100 * (data.number - data.prevNumber) / data.prevNumber; var warning = ''; 
       var currVal = accounting.formatNumber(data.number, options.digits, options.thousand_sep, options.decimal_sep); 
       if(variation > data.thresold1){warning = '<i style="color:'+ options.red +';" title="'+ variation +' increase" class="fas fa-exclamation-triangle"></i>&nbsp;&nbsp;'};
       if(data.url != null){currVal = '<a style="text-decoration:none;color:'+ data.color +'" href="' + data.url+ '">'+ currVal +'</a>'}
       if(data.label != undefined){currVal = data.label +" "+ currVal};
       var progressElt = jHtmlObj.find('#progress');
       progressElt.progressbar({value: variation});
       if(variation > data.thresold1){progressElt.children().css({'background': options.red})} 
       else if(variation > data.thresold2){progressElt.children().css({'background': options.orange})} 
       else{progressElt.children().css({'background': options.green})}
       jHtmlObj.find('div').first().html(warning + currVal);
       jHtmlObj.find('div').last().html('Previous number: '+ accounting.formatNumber(data.prevNumber, options.digits, options.thousand_sep, options.decimal_sep));
      '''

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '''<div %(strAttr)s>
      <div style="width:100%%;text-align:right;font-size:%(size)s"></div>
      <div id="progress" style="height:10px;color:%(color)s;border:1px solid %(greyColor)s"></div>
      <div style="font-size:10px;font-style:italic;color:%(greyColor)s;padding-bottom:5px;text-align:left"></div>
      %(helper)s
      </div>''' % {"strAttr": self.get_attrs(pyClassNames=self.style.get_classes()), "size": Defaults_css.font(12),
                   'htmlId': self.htmlId, "color": self.val['color'],
                   "greyColor": self._report.theme.greys[6], "helper": self.helper}

  # -----------------------------------------------------------------------------------------
  #                                    MARKDOWN SECTION
  # -----------------------------------------------------------------------------------------
  @staticmethod
  def matchMarkDown(val):
    return True if val.startswith("@delta ") else None

  @classmethod
  def convertMarkDown(cls, val, regExpResult, report=None):
    curr, prev = val[6:].split(':')
    if report is not None:
      getattr(report, cls.callFnc)({'number': float(curr), 'prevNumber': float(prev)})
    return ["report.%s( {'number': %s, 'prevNumber': %s} )" % (cls.callFnc, float(curr,), float(prev))]

  @classmethod
  def jsMarkDown(self, vals): return "@delta %s:%s" % (vals['number'], vals['prevNumber'])


class Formula(Html.Html):
  __reqJs = ['mathjs']
  name, category, callFnc = 'Latex Formula', 'Texts', 'formula'

  def __init__(self, report, text, width, color, helper, profile):
    super(Formula, self).__init__(report, text, css_attrs={"color": color, "width": width}, profile=profile)
    self.add_helper(helper)
    #self._report.jsGlobal.addJs("MathJax.Hub.Config({tex2jax: {inlineMath: [['$', '$'], ['\\(', '\\)']]}})")

  @property
  def _js__builder__(self):
    return 'htmlObj.innerHTML = data'

  def __str__(self):
    return '<font %s>%s</font>%s' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.content, self.helper)

  # -----------------------------------------------------------------------------------------
  #                                    MARKDOWN SECTION
  # -----------------------------------------------------------------------------------------
  @staticmethod
  def matchMarkDown(val): return True if val.startswith("$$") and val.strip().endswith("$$") else None

  @classmethod
  def convertMarkDown(cls, val, regExpResult, report):
    if report is not None:
      getattr(report, 'formula')(val.strip())
    return ["report.formula('%s')" % val.strip()]

  @classmethod
  def jsMarkDown(self, vals): return vals


class TrafficLight(Html.Html):
  name, category, callFnc = 'Light', 'Rich', 'light'

  def __init__(self, report, color, label, height, tooltip, helper, profile):
    # Small change to allow the direct use of boolean and none to define the color
    # Those standards will simplify the creation of themes going forward
    super(TrafficLight, self).__init__(report, color, css_attrs={"width": height, "height": height}, profile=profile)
    self.add_helper(helper, css={"margin-top": "-17px"})
    self.add_label(label, css={"width": 'auto', 'float': 'none', 'vertical-align': 'middle', 'height': '100%',
                               "margin": '0 5px', 'display': 'inline-block', "min-width": '100px'})
    self.css({'border-radius': '60px', 'background-color': self.val, 'display': 'inline-block',
              'vertical-align': 'middle'})
    self.set_attrs(name="title", value=tooltip)
    self.set_attrs(name="data-status", value=color)
    self._jsStyles = {'red': self._report.theme.danger[1], 'green': self._report.theme.success[1], 'orange': self._report.theme.warning[1]}
    self.action = None
    if tooltip is not None:
      self.tooltip(tooltip)

  @property
  def dom(self):
    """
    Description:
    ------------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtml.JsHtmlRich
    """
    if self._dom is None:
      self._dom = JsHtml.JsHtmlBackground(self, report=self._report)
    return self._dom

  def colors(self, green=None, red=None, neutral=None):
    """
    Description:
    ------------
    Set the 3 colors of the traffic light

    Attributes:
    ----------
    :param green: The color used in case of result true
    :param red: The color used in case of result false
    :param neutral: The color used in case of null

    :return: self to allow the chains
    """
    if neutral is not None:
      self._jsStyles['orange'] = neutral
    if green is not None:
      self._jsStyles['green'] = green
    if red is not None:
      self._jsStyles['red'] = red
    return self

  def resolve(self, jsFncs, profile=False):
    """
    Description:
    ------------
    Turn a error warning to a green one

    Attributes:
    ----------
    :param jsFncs:
    :param profile:
    """
    self.action = self._report.ui.icon("fas fa-wrench")
    self.action.inReport = False
    self.action.tooltip("Click to try to resolve the issue")
    self.action.style.css.font_size = 8
    self.action.style.css.margin_top = 8
    self.action.style.css.cursor = 'pointer'
    self.action.style.css.vertical_align = 'top'
    self.action.click(jsFncs, profile)
    return self

  def click(self, jsFncs, profile=False):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param jsFncs:
    :param profile:
    """
    success = Colors.getHexToRgb(self._report.theme.success[1])
    self.style.css.cursor = "pointer"
    jsFncs = [self.dom.querySelector("div").toggle("background-color", "rgb(%s, %s, %s)" % (success[0], success[1], success[2]),
                                                   self._report.theme.danger[1])] + jsFncs
    return super(TrafficLight, self).click(jsFncs, profile)

  @property
  def _js__builder__(self):
    return '''
      if(data === false){htmlObj.querySelector('div').style.backgroundColor = options.red}
      else if (data === true){htmlObj.querySelector('div').style.backgroundColor = options.green}
      else if (data === null){htmlObj.querySelector('div').style.backgroundColor = options.orange}
      else {htmlObj.style.backgroundColor = data}'''

  def __str__(self):
    if self.action is not None:
      return '<div id="%s"><div %s></div>%s</div>%s' % (self.htmlId, self.get_attrs(pyClassNames=self.style.get_classes(), withId=False), self.action.html(), self.helper)

    return '<div id="%s"><div %s></div></div>%s' % (self.htmlId, self.get_attrs(pyClassNames=self.style.get_classes(), withId=False), self.helper)

  # -----------------------------------------------------------------------------------------
  #                                    MARKDOWN SECTION
  # -----------------------------------------------------------------------------------------
  @staticmethod
  def matchMarkDown(val): return re.match("-\(\((.*)\)\)-", val)

  @classmethod
  def convertMarkDown(cls, val, regExpResult, report=None):
    if report is not None:
      getattr(report, cls.callFnc)(regExpResult.group(1))
    return ["report.%s('%s')" % (regExpResult.group(1), cls.callFnc)]

  @classmethod
  def jsMarkDown(self, val):
    return "-((%s))-" % val


class ContentsTable(Html.Html):
  name, category, callFnc = 'Contents Table', None, 'contents'

  def __init__(self, report, title, width, height, options, profile):
    self.indices, self.first_level, self.entries_count, self.ext_links = [], None, 0, {}
    super(ContentsTable, self).__init__(report, [], css_attrs={"width": width, "height": height}, profile=profile)
    self.style.css.position = "fixed"
    self.title = self._report.ui.div()
    self.title += self._report.ui.text(title).css({"width": 'auto', 'display': 'inline-block'})
    self.title += self._report.ui.text("[hide]").css({"width": '30px', 'display': 'inline-block', 'margin-left': '5px'})
    self.title[0].style.css.font_size = Defaults_css.font(6)
    self.title[0].style.css.font_weight = "bold"
    self.title.inReport = False

  def __getitem__(self, i):
    """
    Description:
    ------------
    Return the internal column in the row for the given index

    :param i: the column index
    :rtype: Col
    """
    return self.val[i]

  @property
  def style(self):
    """
    Description:
    ------------
    Property to the CSS Style of the component

    :rtype: GrpClsText.ContentTable
    """
    if self._styleObj is None:
      self._styleObj = GrpClsText.ContentTable(self)
    return self._styleObj

  def add(self, text, level, anchor):
    """
    Description:
    ------------

    :param text:
    :param level:
    """
    href = self._report.ui.link(text, url=anchor)
    href.style.css.font_size = Defaults_css.font(2)
    href.style.add_classes.link.no_decoration()
    self.val.append(href)
    href.style.css.display = 'block'
    href.style.css.width = '100%'
    if level is not None:
      href.style.css.padding_left = (level - 1) * 5
    return self

  def move(self):
    super(ContentsTable, self).move()
    self.style.css.position = None
    self.style.css.margin = 5

  def __str__(self):
    div_link = self._report.ui.div(self.val)
    div_link.inReport = False
    self.title[-1].click([div_link.dom.toggle(), self.title[-1].dom.toggleText('[show]', '[hide]')])
    return '''<div %(attr)s>%(title)s%(links)s</div> ''' % {'attr': self.get_attrs(pyClassNames=self.style.get_classes()),
                                                            'title': self.title.html(), 'htmlId': self.htmlId, 'links': div_link.html()}


class SearchResult(Html.Html):
  name, category, callFnc = 'Search Result', 'Text', 'searchr'
  __reqJs = ['jquery']

  def __init__(self, report, recordSet, pageNumber, width, height, options, profile):
    super(SearchResult, self).__init__(report, recordSet, css_attrs={"width": width, "height": height})
    self._jsStyles = {'title': {'color': self._report.theme.colors[7], 'font-size': '18px'}, 'dsc': {'color': self._report.theme.greys[6]},
                      'url': {'color': self._report.theme.success[1], 'font-size': '14px'}, 'visited': {'color': self._report.theme.greys[5]},
                      'link': {'color': self._report.theme.colors[7], 'cursor': 'pointer'}, 'pageNumber': pageNumber,
                      'currPage': 0, "greyColor": self._report.theme.colors[9], "whiteColor": self._report.theme.greys[0]}

  # self.addGlobalFnc("%s(htmlObj, data, jsStyles, currPage)" % self.__class__.__name__, ''' htmlObj.empty() ;
  @property
  def _js__builder__(self):
    return '''
      jHtmlObj = jQuery(htmlObj);
      if (typeof options.currPage == 'undefined'){options.currPage = 0}; var pageNumber = options.pageNumber;
      data.slice(options.currPage * pageNumber).forEach( function(rec){
        var newItem = $('<div style="margin:5px 10px 5px 10px;"></div>') ; 
        var title = $('<div>'+ rec['title'] + '</div>').css( options.title );
        if (rec['urlTitle'] != undefined){
          title.css({'cursor': 'pointer'}); title.click(function(e){window.open(rec['urlTitle'], '_blank')})}
        newItem.append(title);
        if (rec.icon != undefined){
          var item = $('<div></div>').css(options.url);
          item.append( $('<i class="'+ rec['icon'] +'" style="margin-right:5px"></i>')).append(rec['url']);
          newItem.append(item)} 
        else if(rec.url != undefined) {newItem.append($('<div>'+ rec['url'] +'</div>').css(options.url))}
        newItem.append( $('<div>'+ rec['dsc'] +'</div>').css(options.dsc));
        if(rec.visited != undefined){newItem.append($('<div>'+ rec.visited +'</div>').css(options.visited))}
        if(rec.links != undefined){
          rec.links.forEach(function(link){ 
            if (link.url == undefined) {link.url = link.val};
            newItem.append($('<a href='+ link.url +' target="_blank">'+ link.val +'</a><br>').css(options.link))})};
        jHtmlObj.append(newItem);
      }); 
      if(data.length > 0) {
        var reste = data.length/ pageNumber; var currIndex = options.currPage+1; var roundRest = Math.trunc(reste);
        if (roundRest > reste) {reste ++};
        var paginate = $('<div style="display:inline-block;height:35px;padding:0;width:100%%;text-align:center;margin-top:10px" class="py_cssdivpagination"></div>');
        if (currIndex > 1){
          var href = $('<a href="#">&laquo;</a>');
          href.click({page: options.currPage-1, rec: data}, function(e){%(class)s(htmlObj, e.data.rec, options, e.data.page)});
          paginate.append(href)};
        for (var i = 0; i < reste; i++){
          var indexPage = i + 1;
          if (options.currPage == i) { 
            var href = $('<a href="#" style="background-color:'+ options.greyColor +';color:'+ options.whiteColor +'">'+ indexPage +'</a>');
            href.click({page: i, rec: data}, function(e) { %(class)s(htmlObj, e.data.rec, options, e.data.page)});
            paginate.append(href)}
          else{
            var href = $('<a href="#">'+ indexPage +'</a>') ;
            href.click({page: i, rec: data}, function(e){%(class)s(htmlObj, e.data.rec, options, e.data.page)});
            paginate.append(href)}}
        if(currIndex < reste){
          var href = $('<a href="#">&raquo;</a>');
          href.click({page: options.currPage+1, rec: data}, function(e){%(class)s(htmlObj, e.data.rec, options, e.data.page)});
          paginate.append(href)};
        jHtmlObj.append(paginate)
      } ''' % {"class": self.__class__.__name__}

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    #self._report.style.cssCls('CssDivPagination')
    return '<div %s style="margin:5px 10px 5px 10px;"></div> ' % self.get_attrs(pyClassNames=self.style.get_classes())


class Composite(Html.Html):
  name, category, callFnc = 'Composite', 'Rich', 'composite'

  def __init__(self, report, schema, width, height, htmlCode, options, profile, helper):
    super(Composite, self).__init__(report, None, css_attrs={"width": width, "height": height})
    self.__builders, ref_map = set(), {}
    self.__options = OptText.OptionsComposite(self, options)
    self.add_helper(helper)
    self._set_comp(None, schema, self.__builders, ref_map)
    self.attr = self.val.attr
    self._js = self.val._js
    self._dom = self.val._dom
    self._styleObj = self.val._styleObj

  @property
  def options(self):
    """

    :rtype: ptText.OptionsComposite
    """
    return self.__options

  @property
  def dom(self):
    """
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    This is the only flexible component in which this DOM object can be changed.

    :return: A Javascript Dom object

    :rtype: JsHtml.JsHtml
    """
    if self._dom is None:
      self._dom = JsHtml.JsHtml(self.val, report=self._report)
    return self._dom

  @dom.setter
  def dom(self, js_html):
    self._dom = js_html

  @property
  def style(self):
    """
    Description:
    ------------

    :rtype: GrpCls.ClassHtmlEmpty
    """
    if self._styleObj is None:
      self._styleObj = GrpCls.ClassHtmlEmpty(self)
    return self._styleObj

  def __getitem__(self, i):
    return self.val.val[i]

  def __add__(self, htmlObj):
    """ Add items to a container """
    htmlObj.inReport = False # Has to be defined here otherwise it is set to late
    if not isinstance(self.val.val, list):
      self._vals = self.val._vals
    self.val.val.append(htmlObj)
    return self

  @property
  def _get_comp_map(self):
    """
    Description:
    ------------

    """
    return {
      'div': self._report.ui.div,
      'textarea': self._report.ui.textarea,
      'button': self._report.ui.button,
      'label': self._report.ui.label,
      'img': self._report.ui.img,
      'header': self._report.ui.header,
      'section': self._report.ui.section,
      'icon': self._report.ui.icon,
      'span': self._report.ui.texts.span,
      'checkbox': self._report.ui.inputs.checkbox,
      'radio': self._report.ui.inputs.d_radio,
      'input': self._report.ui.inputs.d_text,
      'circle': self._report.ui.charts.svg.circle,
      'svg': self._report.ui.charts.svg.new,
      'path': self._report.ui.charts.svg.path,
      'list': self._report.ui.list,
      'item': self._report.ui.lists.item,
      'nav': self._report.ui.tags.nav,
      'p': self._report.ui.tags.p,
      'link': self._report.ui.tags.a,
      'ol': self._report.ui.tags.ol,
      'aside': self._report.ui.tags.aside,
      'hr': self._report.ui.layouts.hr,
    }

  def _set_comp(self, comp, schema_child, builders, ref_map):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param comp:
    :param schema_child:
    :param builders:
    """
    if 'args' in schema_child and 'url' in schema_child['args']:
      schema_child['args']['url'] = schema_child['args']['url'] % ref_map
    if comp is None:
      # delegate the htmlID to the main component
      new_comp = self._get_comp_map[schema_child['type']](htmlCode=self.htmlId, **schema_child.get('args', {}))
      new_comp.inReport = False
      self._vals = new_comp
    else:
      new_comp = self._get_comp_map[schema_child['type']](**schema_child.get('args', {}))
    if 'builder' in schema_child:
      builders.add(schema_child['builder'])
    if self.options.reset_class and not schema_child.get('class-keep', False):
      new_comp.style.clear()
    if 'class' in schema_child:
      if schema_child['class'] is None:
        new_comp.set_attrs({'class': schema_child['class']})
      else:
        for cls in schema_child['class'].split(" "):
          new_comp.attr['class'].add(cls)
    if 'arias' in schema_child:
      if 'describedby' in schema_child['arias']:
        schema_child['arias']['describedby'] = schema_child['arias']['describedby'] % ref_map
      new_comp.aria.set(schema_child['arias'])
    if 'css' in schema_child:
      new_comp.css(schema_child['css'], reset=True)
    if 'attrs' in schema_child:
      if 'data-target' in schema_child['attrs']:
        schema_child['attrs']['data-target'] = schema_child['attrs']['data-target'] % ref_map
      new_comp.set_attrs(schema_child['attrs'])
    if 'ref' in schema_child:
      ref_map[schema_child['ref']] = new_comp.htmlId
    for child in schema_child.get('children', []):
      self._set_comp(new_comp, child, builders, ref_map)
    if comp is not None:
      comp += new_comp

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).extend(list(self.__builders))
    return self.val.html()




import datetime

# Check if pandas is available in the current environment
# if it is the case this module can handle DataFrame directly
try:
  import pandas as pd
  has_pandas = True

except:
  has_pandas = False


from epyk.core import html


class Fields(object):
  def __init__(self, context):
    self.context = context

  def date(self, value, label=None, icon="far fa-calendar-alt", color=None, htmlCode=None,
            profile=None, options=None, helper=None):
    """
    Description:
    ------------
    This component is based on the Jquery Date Picker object.

    Usage::

      rptObj.ui.fields.date('2020-04-08', label="Date").included_dates(["2020-04-08", "2019-09-06"])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlDates.DatePicker`

    Related Pages:

			https://jqueryui.com/datepicker/

    Attributes:
    ----------
    :param value: Optional. The value to be displayed to the time component. Default now
    :param label: Optional. The text of label to be added to the component
    :param icon: Optional. The component icon content from font-awesome references
    :param color: Optional. The font color in the component. Default inherit
    :param htmlCode: Optional. An identifier for this component (on both Python and Javascript side)
    :param profile: Optional. A flag to set the component performance storage
    :param options: Optional. Specific Python options available for this component
    :param helper: Optional. A tooltip helper
    """
    html_dt = html.HtmlDates.DatePicker(self.context.rptObj, value, label, icon, color, htmlCode, profile, options or {}, helper)
    self.context.register(html_dt)
    return html_dt

  def today(self, label=None, icon="far fa-calendar-alt", color=None, htmlCode=None,
            profile=None, options=None, helper=None):
    """
    Description:
    ------------
    This component is based on the Jquery Date Picker object.

    Usage::

      rptObj.ui.fields.today(label="Date").included_dates(["2019-09-01", "2019-09-06"])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlDates.DatePicker`

    Related Pages:

			https://jqueryui.com/datepicker/

    Attributes:
    ----------
    :param label: Optional. The text of label to be added to the component
    :param icon: Optional. The component icon content from font-awesome references
    :param color: Optional. The font color in the component. Default inherit
    :param htmlCode: Optional. An identifier for this component (on both Python and Javascript side)
    :param profile: Optional. A flag to set the component performance storage
    :param options: Optional. Specific Python options available for this component
    :param helper: Optional. A tooltip helper
    """
    value = self.context.rptObj.py.dates.today
    html_dt = html.HtmlDates.DatePicker(self.context.rptObj, value, label, icon, color, htmlCode, profile, options or {}, helper)
    self.context.register(html_dt)
    return html_dt

  def cob(self, label=None, icon="far fa-calendar-alt", color=None, htmlCode=None,
          profile=None, options=None, helper=None):
    """
    Description:
    ------------
    This component is based on the Jquery Date Picker object.

    Usage::

      rptObj.ui.fields.cob(label="COB Date")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlDates.DatePicker`

    Related Pages:

			https://jqueryui.com/datepicker/

    Attributes:
    ----------
    :param label: Optional. The text of label to be added to the component
    :param icon: Optional. The component icon content from font-awesome references
    :param color: Optional. The font color in the component. Default inherit
    :param htmlCode: Optional. An identifier for this component (on both Python and Javascript side)
    :param profile: Optional. A flag to set the component performance storage
    :param options: Optional. Specific Python options available for this component
    :param filters: Optional. The filtering properties for this component
    :param helper: Optional. A tooltip helper
    """
    value = self.context.rptObj.py.dates.cob
    html_cob = html.HtmlDates.DatePicker(self.context.rptObj, value, label, icon, color, htmlCode, profile, options or {}, helper)
    self.context.register(html_cob)
    return html_cob

  def now(self, deltatime=0, label=None, icon="far fa-clock", color=None, htmlCode=None, profile=None,
          options=None, helper=None):
    """
    Description:
    ------------
    This component is based on the Jquery Time Picker object.

    Usage::

      rptObj.ui.fields.now(label="timestamp", color="red", helper="This is the report timestamp")
      rptObj.ui.fields.now(label="Time field")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlDates.TimePicker`

    Related Pages:

			https://github.com/jonthornton/jquery-timepicker

    Attributes:
    ----------
    :param label: Optional. The text of label to be added to the component
    :param icon: Optional. The component icon content from font-awesome references
    :param color: Optional. The font color in the component. Default inherit
    :param htmlCode: Optional. An identifier for this component (on both Python and Javascript side)
    :param profile: Optional. A flag to set the component performance storage
    :param options: Optional. Specific Python options available for this component
    :param helper: Optional. A tooltip helper
    """
    date = datetime.datetime.now() + datetime.timedelta(minutes=deltatime)
    html_dt = html.HtmlDates.TimePicker(self.context.rptObj, str(date).split(" ")[1].split(".")[0], label, icon, color,
                                        htmlCode, profile, options or {}, helper)
    self.context.register(html_dt)
    return html_dt

  def time(self, value=None, label=None, icon="far fa-clock", color=None, htmlCode=None, profile=None,
          options=None, helper=None):
    """
    Description:
    ------------
    This component is based on the Jquery Time Picker object.

    Usage::

      rptObj.ui.fields.time(label="timestamp", color="red", helper="This is the report timestamp")
      rptObj.ui.fields.time(label="Time field")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlDates.TimePicker`

    Related Pages:

			https://github.com/jonthornton/jquery-timepicker

    Attributes:
    ----------
    :param value: Optional. The value to be displayed to the time component. Default now
    :param label: Optional. The text of label to be added to the component
    :param icon: Optional. The component icon content from font-awesome references
    :param color: Optional. The font color in the component. Default inherit
    :param htmlCode: Optional. An identifier for this component (on both Python and Javascript side)
    :param profile: Optional. A flag to set the component performance storage
    :param options: Optional. Specific Python options available for this component
    :param helper: Optional. A tooltip helper
    """
    html_dt = html.HtmlDates.TimePicker(self.context.rptObj, value, label, icon, color, htmlCode, profile, options or {}, helper)
    self.context.register(html_dt)
    return html_dt

  def input(self, value="", label=None, placeholder="", icon=None, width=(100, "%"),
            height=(None, "px"), htmlCode=None, helper=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.fields.input("", label="Range Example", icon="fas fa-unlock-alt")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.FieldInput`

    Attributes:
    ----------
    :param value:
    :param label:
    :param placeholder:
    :param icon:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    html_input = html.HtmlInput.FieldInput(self.context.rptObj, value, label, placeholder, icon, width, height, htmlCode, helper, profile)
    self.context.register(html_input)
    return html_input

  def autocomplete(self, value="", label=None, placeholder="", icon=None, width=(100, "%"),
            height=(None, "px"), htmlCode=None, helper=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.fields.input("", label="Range Example", icon="fas fa-unlock-alt")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.FieldAutocomplete`

    Attributes:
    ----------
    :param value:
    :param label:
    :param placeholder:
    :param icon:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    html_input = html.HtmlInput.FieldAutocomplete(self.context.rptObj, value, label, placeholder, icon, width, height, htmlCode, helper, profile)
    self.context.register(html_input)
    return html_input

  def static(self, value="", label=None, placeholder="", icon=None, width=(100, "%"), height=(None, "px"), htmlCode=None,
             helper=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.fields.static(label="readonly field")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.FieldInput`

    Attributes:
    ----------
    :param value:
    :param label:
    :param placeholder:
    :param icon:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    html_input = html.HtmlInput.FieldInput(self.context.rptObj, value, label, placeholder, icon, width, height, htmlCode, helper, profile)
    html_input.input.readonly(True)
    self.context.register(html_input)
    return html_input

  def integer(self, value="", label=None, placeholder="", icon=None, width=(100, "%"),
              height=(None, "px"), htmlCode=None, helper=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.fields.integer(label="test")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.FieldInteger`

    Attributes:
    ----------
    :param value:
    :param label:
    :param placeholder:
    :param icon:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    html_input = html.HtmlInput.FieldInteger(self.context.rptObj, value, label, placeholder, icon, width, height, htmlCode, helper, profile)
    self.context.register(html_input)
    return html_input

  def password(self, value="", label=None, placeholder="", icon=None, width=(100, "%"),
              height=(None, "px"), htmlCode=None, helper=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.fields.password(label="password")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.FieldPassword`

    Attributes:
    ----------
    :param value:
    :param label:
    :param placeholder:
    :param icon:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    html_input = html.HtmlInput.FieldPassword(self.context.rptObj, value, label, placeholder, icon, width, height,
                                              htmlCode, helper, profile)
    self.context.register(html_input)
    return html_input

  def textarea(self, value="", label=None, placeholder="", icon=None, width=(100, "%"),
              height=(None, "px"), htmlCode=None, helper=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.fields.textarea(label="Date")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.FieldTextArea`

    Attributes:
    ----------
    :param value:
    :param label:
    :param placeholder:
    :param icon:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    html_input = html.HtmlInput.FieldTextArea(self.context.rptObj, value, label, placeholder, icon, width, height,
                                              htmlCode, helper, profile)
    self.context.register(html_input)
    return html_input

  def checkbox(self, value=False, label=None, icon=None, width=(100, "%"), height=(None, "px"),
               htmlCode=None, helper=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.fields.checkbox(True, label="Check")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.FieldCheckBox`

    Related Pages:

			https://www.w3schools.com/tags/att_input_type_checkbox.asp

    Attributes:
    ----------
    :param value:
    :param label:
    :param icon:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    html_input = html.HtmlInput.FieldCheckBox(self.context.rptObj, value, label, icon, width, height, htmlCode, helper, profile)
    self.context.register(html_input)
    return html_input

  def radio(self, value=False, label=None, group_name=None, icon=None, width=(100, "%"),
              height=(None, "px"), htmlCode=None, helper=None, profile=None):
    """
    Description:
    ------------
    The <input type="radio"> defines a radio button.
    Radio buttons are normally presented in radio groups (a collection of radio buttons describing a set of related options).
    Only one radio button in a group can be selected at the same time.

    Usage::

      rptObj.ui.inputs.radio(False, label="radio")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.Radio`

    Related Pages:

			https://www.w3schools.com/tags/att_input_type_radio.asp

    Attributes:
    ----------
    :param value:
    :param label:
    :param group_name:
    :param icon:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    html_input = html.HtmlInput.Radio(self.context.rptObj, value, label, group_name, icon, width, height, htmlCode, helper, profile)
    html_input.label.css({"width": '100px', 'float': 'left'})
    html_input.css({"display": 'inline-block'})
    self.context.register(html_input)
    return html_input

  def range(self, value="", min=0, max=100, step=1, label=None, placeholder="", icon=None, width=(100, "%"),
              height=(None, "px"), htmlCode=None, helper=None, options=None, profile=None):
    """
    Description:
    ------------
    The <input type="range"> defines a control for entering a number whose exact value is not important (like a slider control).
    Default range is 0 to 100. However, you can set restrictions on what numbers are accepted with the attributes below.
    - max - specifies the maximum value allowed
    - min - specifies the minimum value allowed
    - step - specifies the legal number intervals
    - value - Specifies the default value

    Usage::

      rptObj.ui.fields.range(54, min=20, label="Range Example", icon="fas fa-unlock-alt")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.FieldRange`

    Related Pages:

			https://www.w3schools.com/tags/att_input_type_range.asp

    Attributes:
    ----------
    :param value:
    :param min:
    :param max:
    :param step:
    :param label:
    :param placeholder:
    :param icon:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    html_input = html.HtmlInput.FieldRange(self.context.rptObj, value, min, max, step, label, placeholder, icon, width,
                                           height, htmlCode, helper, options, profile)
    self.context.register(html_input)
    return html_input

  def select(self, value=False, label=None, icon=None, width=(100, "%"), height=(None, "px"), htmlCode=None,
             helper=None, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.fields.select(["a", "b"], label="Check")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlInput.FieldSelect`

    Attributes:
    ----------
    :param value:
    :param label:
    :param icon:
    :param width:
    :param height:
    :param htmlCode:
    :param profile:
    """
    if options is not None and 'align' in options:
      self.context.rptObj.css.customText('.filter-option-inner-inner {text-align: %s}' % options['align'])
    html_input = html.HtmlInput.FieldSelect(self.context.rptObj, value, label, icon, width, height, htmlCode, helper, profile)
    html_input.input.attr['data-width'] = '%spx' % html.Defaults.INPUTS_MIN_WIDTH
    self.context.register(html_input)
    return html_input



"""
Module dedicated to perform the data transformation for the Plotly charts
"""

import json

from epyk.core.js import JsUtils

# List of defined names for the country charts
CTY = ['Belarus', 'Moldova', 'Lithuania', 'Russia', 'Romania', 'Ukraine', 'Andorra', 'Hungary', 'Czech Republic',
       'Slovakia', 'Portugal', 'Serbia', 'Grenada', 'Poland', 'Latvia', 'Finland', 'South Korea', 'France', 'Australia',
       'Croatia', 'Ireland', 'Luxembourg', 'Germany', 'Slovenia', 'United Kingdom', 'Denmark', 'Bulgaria', 'Spain',
       'Belgium', 'South Africa', 'New Zealand', 'Gabon', 'Namibia', 'Switzerland', 'Saint Lucia', 'Austria', 'Estonia',
       'Greece', 'Kazakhstan', 'Canada', 'Nigeria', 'Netherlands', 'Uganda', 'Rwanda', 'Chile', 'Argentina', 'Burundi',
       'United States', 'Cyprus', 'Sweden', 'Venezuela', 'Paraguay', 'Brazil', 'Sierra Leone', 'Montenegro', 'Belize',
       'Cameroon', 'Botswana', 'Saint Kitts and Nevis', 'Guyana', 'Peru', 'Panama', 'Niue', 'Palau', 'Norway', 'Tanzania',
       'Georgia', 'Uruguay', 'Angola', 'Laos', 'Japan', 'Mexico', 'Ecuador', 'Dominica', 'Iceland', 'Thailand',
       'Bosnia and Herzegovina', 'Sao Tome and Principe', 'Malta', 'Albania', 'Bahamas', 'Dominican Republic',
       'Mongolia', 'Cape Verde', 'Barbados', 'Burkina Faso', 'Italy', 'Trinidad and Tobago', 'China', 'Macedonia',
       'Saint Vincent and the Grenadines', 'Equatorial Guinea', 'Suriname', 'Vietnam', 'Lesotho', 'Haiti', 'Cook Islands',
       'Colombia', 'Ivory Coast', 'Bolivia', 'Swaziland', 'Zimbabwe', 'Seychelles', 'Cambodia', 'Puerto Rico',
       'Netherlands Antilles', 'Philippines', 'Costa Rica', 'Armenia', 'Cuba', 'Nicaragua', 'Jamaica', 'Ghana', 'Liberia',
       'Uzbekistan', 'Chad', 'United Arab Emirates', 'Kyrgyzstan', 'India', 'Turkmenistan', 'Kenya', 'Ethiopia', 'Honduras',
       'Guinea-Bissau', 'Zambia', 'Republic of the Congo', 'Guatemala', 'Central African Republic', 'North Korea',
       'Sri Lanka', 'Mauritius', 'Samoa', 'Democratic Republic of the Congo', 'Nauru', 'Gambia',
       'Federated States of Micronesia', 'El Salvador', 'Fiji', 'Papua New Guinea', 'Kiribati', 'Tajikistan',
       'Israel', 'Sudan', 'Malawi', 'Lebanon', 'Azerbaijan', 'Mozambique', 'Togo', 'Nepal', 'Brunei', 'Benin',
       'Singapore', 'Turkey', 'Madagascar', 'Solomon Islands', 'Tonga', 'Tunisia', 'Tuvalu', 'Qatar', 'Vanuatu',
       'Djibouti', 'Malaysia', 'Syria', 'Maldives', 'Mali', 'Eritrea', 'Algeria', 'Iran', 'Oman', 'Brunei', 'Morocco',
       'Jordan', 'Bhutan', 'Guinea', 'Burma', 'Afghanistan', 'Senegal', 'Indonesia', 'Timor-Leste', 'Iraq', 'Somalia',
       'Egypt', 'Niger', 'Yemen', 'Comoros', 'Saudi Arabia', 'Bangladesh', 'Kuwait', 'Libya', 'Mauritania', 'Pakistan']


class JsChartPlotlyLinks(object):
  def __init__(self, data, js_src, data_schema=None, profile=False):
    self._js_src, self._data_schema, self.profile, self._data = js_src, data_schema, profile, data

  def __register_records_fnc(self, fnc_name, fnc_def, fnc_pmts=None, profile=False):
    """
    This function will attach to the report object only the javascript functions used during the report

    :param fnc_name: A String with the Javascript function name to be defined
    :param fnc_def: A String with the Javascript function content
    :param fnc_pmts: A list of parameters
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    fnc_pmts = ["data"] + (fnc_pmts or [])
    if not fnc_name in self._js_src.get('js', {}).get('functions', {}):
      if profile:
        content = "var result = []; %s;return result" % JsUtils.cleanFncs(fnc_def)
      else:
        content = "var result = []; %s;return result" % JsUtils.cleanFncs(fnc_def)
      self._js_src.setdefault('js', {}).setdefault('functions', {})[fnc_name] = {'content': content, 'pmt': fnc_pmts}

  def line(self, y_columns, x_axis, profile=False):
    """
    Data Conversion to a Plotly Line chart object.

    This entry point will register the Javascript function and then use it on the javascript side using the parameters
    passed in this function. Some basic controls are done on the Python side during the transpilation

    Example

    :param y_columns: A python list with the columns in the records
    :param x_axis: A string with a column name from the records
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    if not isinstance(y_columns, list):
      y_columns = [y_columns]
    fnc_name = JsPlotly.__name__
    x_axis = JsUtils.jsConvertData(x_axis, None)
    self.__register_records_fnc(fnc_name, JsPlotly.content, list(JsPlotly.pmts), profile | self.profile)
    self._data_schema['out'] = "%s(%%s, %s, %s)" % (fnc_name, y_columns, x_axis)

  def pie(self, y_columns, x_axis, profile=False):
    """
    Data Conversion to a Plotly Pie chart object.

    This entry point will register the Javascript function and then use it on the javascript side using the parameters
    passed in this function. Some basic controls are done on the Python side during the transpilation

    Example

    :param y_columns: A python list with the columns in the records
    :param x_axis: A string with a column name from the records
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    if not isinstance(y_columns, list):
      y_columns = [y_columns]
    fnc_name = JsPlotlyPie.__name__
    x_axis = JsUtils.jsConvertData(x_axis, None)
    self.__register_records_fnc(fnc_name, JsPlotlyPie.content, list(JsPlotlyPie.pmts), profile | self.profile)
    self._data_schema['out'] = "%s(%%s, %s, %s)" % (fnc_name, y_columns, x_axis)

  def heatmap(self, y_columns, x_axis, profile=False):
    """
    Data Conversion to a Plotly Heatmap chart object.

    This entry point will register the Javascript function and then use it on the javascript side using the parameters
    passed in this function. Some basic controls are done on the Python side during the transpilation

    Example

    :param y_columns: A python list with the columns in the records
    :param x_axis: A string with a column name from the records
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    if not isinstance(y_columns, list):
      y_columns = [y_columns]
    fnc_name = JsPlotlyHeatmap.__name__
    x_axis = JsUtils.jsConvertData(x_axis, None)
    self.__register_records_fnc(fnc_name, JsPlotlyHeatmap.content, list(JsPlotlyHeatmap.pmts), profile | self.profile)
    self._data_schema['out'] = "%s(%%s, %s, %s)" % (fnc_name, y_columns, x_axis)

  def map(self, y_columns, x_axis, profile=False):
    """
    Data Conversion to a Plotly Map chart object.

    This entry point will register the Javascript function and then use it on the javascript side using the parameters
    passed in this function. Some basic controls are done on the Python side during the transpilation

    Example

    :param y_columns: A python list with the columns in the records
    :param x_axis: A string with a column name from the records
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    if not isinstance(y_columns, list):
      y_columns = [y_columns]
    fnc_name = JsPlotlyMaps.__name__
    x_axis = JsUtils.jsConvertData(x_axis, None)
    self.__register_records_fnc(fnc_name, JsPlotlyMaps.content, list(JsPlotlyMaps.pmts), profile | self.profile)
    self._data_schema['out'] = "%s(%%s, %s, %s)" % (fnc_name, y_columns, x_axis)

  def scatter(self, y_columns, x_axis, profile=False):
    """
    Data Conversion to a Plotly Scatter chart object.

    This entry point will register the Javascript function and then use it on the javascript side using the parameters
    passed in this function. Some basic controls are done on the Python side during the transpilation

    Example

    :param y_columns: A python list with the columns in the records
    :param x_axis: A string with a column name from the records
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    if not isinstance(y_columns, list):
      y_columns = [y_columns]
    fnc_name = JsPlotlyScatter.__name__
    x_axis = JsUtils.jsConvertData(x_axis, None)
    self.__register_records_fnc(fnc_name, JsPlotlyScatter.content, list(JsPlotlyScatter.pmts), profile | self.profile)
    self._data_schema['out'] = "%s(%%s, %s, %s)" % (fnc_name, y_columns, x_axis)

  def box(self, y_columns, profile=False):
    """
    Data Conversion to a Plotly Box chart object.

    This entry point will register the Javascript function and then use it on the javascript side using the parameters
    passed in this function. Some basic controls are done on the Python side during the transpilation

    Example

    :param y_columns: A python list with the columns in the records
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    if not isinstance(y_columns, list):
      y_columns = [y_columns]
    fnc_name = JsPlotlyBox.__name__
    self.__register_records_fnc(fnc_name, JsPlotlyBox.content, list(JsPlotlyBox.pmts), profile | self.profile)
    self._data_schema['out'] = "%s(%%s, %s)" % (fnc_name, y_columns)

  def h_box(self, y_columns, profile=False):
    """
    Data Conversion to a Plotly Horizontal Box chart object.

    This entry point will register the Javascript function and then use it on the javascript side using the parameters
    passed in this function. Some basic controls are done on the Python side during the transpilation

    Example

    :param y_columns: A python list with the columns in the records
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    if not isinstance(y_columns, list):
      y_columns = [y_columns]
    fnc_name = JsPlotlyHBox.__name__
    self.__register_records_fnc(fnc_name, JsPlotlyHBox.content, list(JsPlotlyHBox.pmts), profile | self.profile)
    self._data_schema['out'] = "%s(%%s, %s)" % (fnc_name, y_columns)

  def h_bar(self, y_columns, profile=False):
    """
    Data Conversion to a Plotly Horizontal Bar chart object.

    This entry point will register the Javascript function and then use it on the javascript side using the parameters
    passed in this function. Some basic controls are done on the Python side during the transpilation

    Example

    :param y_columns: A python list with the columns in the records
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    if not isinstance(y_columns, list):
      y_columns = [y_columns]
    fnc_name = JsPlotlyHBar.__name__
    self.__register_records_fnc(fnc_name, JsPlotlyHBar.content, list(JsPlotlyHBar.pmts), profile | self.profile)
    self._data_schema['out'] = "%s(%%s, %s)" % (fnc_name, y_columns)

  def surface(self, y_columns, x_axis, z_axis, profile=False):
    """
    Data Conversion to a Plotly Surface chart object.

    This entry point will register the Javascript function and then use it on the javascript side using the parameters
    passed in this function. Some basic controls are done on the Python side during the transpilation

    Example

    :param y_columns: A python list with the columns in the records
    :param x_axis: A string with a column name from the records
    :param z_axis: A string with a column name from the records
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    if not isinstance(y_columns, list):
      y_columns = [y_columns]
    fnc_name = JsPlotlySurface.__name__
    x_axis = JsUtils.jsConvertData(x_axis, None)
    self.__register_records_fnc(fnc_name, JsPlotlySurface.content, list(JsPlotlySurface.pmts), profile | self.profile)
    self._data_schema['out'] = "%s(%%s, %s, %s, %s)" % (fnc_name, y_columns, x_axis, JsUtils.jsConvertData(z_axis, None))

  def parcoords(self, y_columns, x_axis, profile=False):
    """
    Data Conversion to a Plotly ParCoordinate chart object.

    This entry point will register the Javascript function and then use it on the javascript side using the parameters
    passed in this function. Some basic controls are done on the Python side during the transpilation

    Example

    :param y_columns: A python list with the columns in the records
    :param x_axis: A string with a column name from the records
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    if not isinstance(y_columns, list):
      y_columns = [y_columns]
    fnc_name = JsPlotlyParCoordinates.__name__
    x_axis = JsUtils.jsConvertData(x_axis, None)
    self.__register_records_fnc(fnc_name, JsPlotlyParCoordinates.content, list(JsPlotlyParCoordinates.pmts), profile | self.profile)
    self._data_schema['out'] = "%s(%%s, %s, %s)" % (fnc_name, y_columns, x_axis)


# --------------------------------------------------------------------------------------------------------------------
#                               DATA TRANSFORMATION JAVASCRIPT DEFINITION
# --------------------------------------------------------------------------------------------------------------------
class JsPlotlyPie(object):
  pmts = ("seriesNames", "xAxis")
  content = '''
    var temp = {}; var labels = {};
    data.forEach(function(rec){ 
      if (!(rec[xAxis] in temp)){temp[rec[xAxis]] = {}};
      seriesNames.forEach(function(name){
        labels[name] = true; if(rec[name] !== undefined){
          if(!(name in temp[rec[xAxis]])){temp[rec[xAxis]][name] = rec[name]} else {temp[rec[xAxis]][name] += rec[name]}}})
    });
    result = [{labels: [], values: []}]; var labels = Object.keys(labels);
    for(var series in temp){
      labels.forEach(function(label){if(temp[series][label] !== undefined){
        result[0].labels.push(series); result[0].values.push(temp[series][label])}})}
    '''


class JsPlotly(object):
  pmts = ("seriesNames", "xAxis")
  content = '''
    var temp = {}; var labels = []; var uniqLabels = {};
    seriesNames.forEach(function(series){temp[series] = {}});
    data.forEach(function(rec){ 
      seriesNames.forEach(function(name){
        if(rec[name] !== undefined){
          if(!(rec[xAxis] in uniqLabels)){labels.push(rec[xAxis]); uniqLabels[rec[xAxis]] = true};
          temp[name][rec[xAxis]] = rec[name]}})});
    seriesNames.forEach(function(series){
      dataSet = {x: [], y: [], name: series};
      labels.forEach(function(x, i){
        dataSet.x.push(x);
        if(temp[series][x] == undefined){dataSet.y.push(null)} else{dataSet.y.push(temp[series][x])}
      }); result.push(dataSet)})
    '''


class JsPlotlyScatter(object):
  pmts = ("seriesNames", "xAxis")
  content = '''
    var temp = {}; var labels = [];
    seriesNames.forEach(function(series){temp[series] = {x: [], y: [], name: series}});
    data.forEach(function(rec){ 
      seriesNames.forEach(function(name){
        if(rec[name] !== undefined){temp[name].x.push(rec[xAxis]); temp[name].y.push(rec[name])}})});
    seriesNames.forEach(function(series){result.push(temp[series])})
    '''


class JsPlotlyBox(object):
  pmts = ("seriesNames", )
  content = '''
    var temp = {}; seriesNames.forEach(function(name){temp[name] = []});
    data.forEach(function(rec){ 
      seriesNames.forEach(function(name){if (rec[name] !== undefined){temp[name].push(rec[name])}})}); 
    for(var name in temp){result.push({type: 'box', name: name, y: temp[name]})}
    '''


class JsPlotlyHBox(object):
  pmts = ("seriesNames", )
  content = '''
    var temp = {}; seriesNames.forEach(function(name){temp[name] = []});
    data.forEach(function(rec){ 
      seriesNames.forEach(function(name){if(rec[name] !== undefined){temp[name].push(rec[name])}})}); 
    for(var name in temp){result.push({type: 'box', name: name, x: temp[name]})}
    '''


class JsPlotlySurface(object):
  pmts = ("seriesNames", "xAxis", "zAxis")
  content = '''
    seriesNames.forEach(function(series){
      records = {z: []};
      data.forEach(function(rec){var y = rec[series]; if(y == undefined){y = 0};
         records.z.push([parseFloat(rec[xAxis]), y, rec[zAxis]])})
      result.push(records)})
    '''


class JsPlotlyScatter3D(object):
  """
  Convert the recordset to a Plotly object which can be used in 3d surface charts

  Example
  report.chart('surface-contours', df1, seriesNames=['col2'], xAxis='col4', otherDims=['col3'])
  report.chart('mesh3d', df1, seriesNames=['col2', 'col3'], xAxis='col4', otherDims=['col5'], height=600)
  """
  alias = "Plotly"
  chartTypes = ['scatter3d', 'mesh3d']
  params = ("seriesNames", "xAxis", "zAxis")
  value = '''
    seriesNames.forEach(function(s){result.push({x: [], y: [], z:[], name: s})});
    data.forEach(function(rec){
      seriesNames.forEach(function(s, i){
        var y = rec[s]; if(y == undefined) {y = 0};
        result[i].x.push(rec[xAxis]); result[i].y.push(y); result[i].z.push(rec[zAxis])})})
    '''


class JsPlotlyHBar(object):
  pmts = ("seriesNames", "xAxis")
  content = '''
    var temp = {}; var labels = {};
    data.forEach(function(rec){ 
      if(!(rec[xAxis] in temp)){temp[rec[xAxis]] = {}};
      seriesNames.forEach(function(name){
        labels[name] = true; if(rec[name] !== undefined){
          if(!(name in temp[rec[xAxis]])){temp[rec[xAxis]][name] = rec[name]}else{temp[rec[xAxis]][name] += rec[name]}}})
    });
    result = []; var labels = Object.keys(labels);
    labels.forEach(function(label){
      dataSet = {x: [], y: [], name: label};
      for(var series in temp){if(temp[series][label] !== undefined){dataSet.x.push(series); dataSet.y.push(temp[series][label])}};
      result.push(dataSet)}) 
    '''


class JsPlotlyMaps(object):
  """
  Transform a recordset to a data structure usable for a Plotly map chart.
  The DataFrame should have the below structure:

    index col2  col5
    0     2     France
    1     5     Russia

  Example
  report.chart('europe', df1, seriesNames=['col2'], xAxis='col5')
  report.chart('choropleth', df1, seriesNames=['col2'], xAxis='col5')
  """
  pmts = ("seriesNames", "xAxis")
  content = '''
    var countries = %(countries)s; var dataSet = {locations:[], z: [], text: [], autocolorscale: true}; var temp = {};
    data.forEach(function(rec){ 
      seriesNames.forEach(function(name){
        if(temp[rec[xAxis]] == undefined){temp[rec[xAxis]] = rec[name]} else{temp[rec[xAxis]] += rec[name]}})});
    countries.forEach(function(country){
      dataSet.locations.push(country); dataSet.text.push(country); 
      if(country in temp){dataSet.z.push(temp[country])} else {dataSet.z.push(0)}});
    result.push(dataSet)
    ''' % {'countries': json.dumps(CTY)}


class JsPlotlyHeatmap(object):
  pmts = ("seriesNames", "xAxis")
  content = '''
    result = [{x: [], y: [], z: []}];
    seriesNames.forEach(function(s){result[0].x.push(s)});
    data.forEach(function(rec){
      result[0].y.push(rec[xAxis]); var zDim = []; seriesNames.forEach(function(s){ zDim.push(rec[s])});
      result[0].z.push(zDim)}) 
    ''' % {'countries': json.dumps(CTY)}


class JsPlotlyParCoordinates(object):
  pmts = ("seriesNames", "xAxis")
  content = '''
    dimensions = []; var counts = [];
    seriesNames.forEach(function(series){dimensions.push({label: series, values: []})});
    data.forEach(function(rec){ 
      counts.push(parseInt(rec[xAxis])); 
      seriesNames.forEach(function(series, i){
        if(rec[series] == undefined){dimensions[i].values.push(0)} else {dimensions[i].values.push(rec[series])}})});
    result = [{dimensions: dimensions, counts: counts}]
    '''




from epyk.core.css.styles.classes import CssStyle


class CssDropDownSubMenu(CssStyle.Style):
  _attrs = {'display': 'none'}
  classname = "menu .submenu"


class CssDropDownMenuHLi(CssStyle.Style):
  _attrs = {'display': 'block'}
  classname = "menu li.dropdown:hover > .submenu"


class CssDropDownMenuHoverAAfter(CssStyle.Style):
  _attrs = {'display': 'block'}
  classname = "menu ul.submenu > li.dropdown:hover > .submenu"


class CssDropDownSubMenuPullLeft(CssStyle.Style):
  _attrs = {'position': 'absolute', 'top': '0px', 'list-style-type': 'none', 'white-space': 'nowrap', 'margin': 0, 'padding': 0} # 'left': '0px',
  classname = "menu .submenu .submenu"


class CssDropDownMenu(CssStyle.Style):
  _attrs = {'position': 'absolute', 'top': '0px', 'list-style-type': 'none', 'white-space': 'nowrap', 'margin': 0, 'padding': 0} # 'left': '0px',
  classname = "menu ul.submenu"


class CssDropDownAfterMenu(CssStyle.Style):
  _attrs = {'display': 'block'}
  classname = "menu li a"

  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1], 'background': self.rptObj.theme.greys[0]})



from . import PyExt
from . import PyOuts


class OrderedSet(list):
  def __init__(self):
    super(OrderedSet, self).__init__()

  def discard(self, value):
    """

    :param value:
    """
    self.remove(value)

  def add(self, key):
    if key not in self:
      self.append(key)




from epyk.core.html import Html

from epyk.core.html import Defaults
from epyk.core.html.options import OptPanel
from epyk.core.css.styles import GrpClsContainer
from epyk.core.js.html import JsHtmlStepper


class Drawer(Html.Html):

  def __init__(self, report, width, height, options, helper, profile):
    super(Drawer, self).__init__(report, None, css_attrs={"width": width, "height": height})
    self.add_helper(helper, css={"line-height": '%spx' % Defaults.LINE_HEIGHT})
    self.__options = OptPanel.OptionDrawer(self, options)
    self.style.css.position = 'relative'

    self.panels = report.ui.div()
    self.panels.inReport = False
    self.panels.attr['name'] = 'drawer_panels'

    self.handle = report.ui.div()
    self.handle.style.clear_all()
    self.handle.style.css.cursor = 'pointer'

    self.handle.inReport = False
    self.handle.attr['name'] = 'drawer_handle'

    self.drawers = report.ui.div()
    self.drawers.style.clear_all()
    self.drawers.inReport = False
    self.drawers.attr['name'] = 'drawer_content'

  @property
  def dom(self):
    """
    Description:
    ------------

    :rtype: JsHtmlStepper.Drawer
    """
    if self._dom is None:
      self._dom = JsHtmlStepper.Drawer(self, report=self._report)
    return self._dom

  @property
  def options(self):
    """
    Description:
    ------------
    Property to set all the possible object for a drawer

    :rtype: OptPanel.OptionDrawer
    """
    return self.__options

  def add_panel(self, link, container):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param link:
    :param container:
    """
    if not hasattr(link, 'inReport'):
      link = self._report.ui.div(link)
      link.inReport = False
    if not hasattr(container, 'inReport'):
      container = self._report.ui.div(container)
      container.style.css.display = 'none'
      container.inReport = False
    self.panels += container
    self.drawers += link

  @property
  def style(self):
    """
    Description:
    ------------

    :rtype: GrpClsContainer.ClassDrawer
    """
    if self._styleObj is None:
      self._styleObj = GrpClsContainer.ClassDrawer(self)
    return self._styleObj

  def set_handle(self, component):
    """
    Description:
    ------------

    """
    self.handle = self._report.ui.div()
    self.handle.style.clear_all()
    if self.options.side == 'left':
      component.click([self.drawers.dom.toggle_transition("margin-right", "0px", "-%s" % self.options.width)])
    else:
      component.click([self.drawers.dom.toggle_transition("margin-left", "0px", "-%s" % self.options.width)])

  def __str__(self):
    self.handle.style.css.float = self.options.side
    if self.options.side == 'left':
      self.drawers.style.css.width = self.options.width
      self.drawers.style.css.margin_right = "-%s" % self.options.width
      self.handle.click([self.drawers.dom.toggle_transition("margin-right", "0px", "-%s" % self.options.width)])
    else:
      self.drawers.style.css.width = self.options.width
      self.drawers.style.css.margin_left = "-%s" % self.options.width
      self.handle.click([self.drawers.dom.toggle_transition("margin-left", "0px", "-%s" % self.options.width)])
    position = {"left": 'right', 'right': 'left'}
    return '''
      <div %(attr)s>
        %(panels)s
        <div name='drawer' style='%(side)s:0'>
          %(handle)s%(drawer)s
        </div>
      </div>''' % {'attr': self.get_attrs(pyClassNames=self.style.get_classes()), 'htmlId': self.htmlId,
                   'drawer': self.drawers.html(), 'handle': self.handle.html(), 'panels': self.panels.html(), 'side': position[self.options.side]}



"""
All those components are a bit special as they would require to work properly:
  - A server in order to interact dynamically
  - A database source

This can definition be configured but by default it is using the ones of the server.
"""

import json
import datetime

from epyk.core.html import Html

# The list of CSS classes
# from epyk.core.css.styles import CssGrpClsText

from epyk.core.js.Imports import requires
sqlalchemy = requires("sqlalchemy", reason='Missing Package', install='sqlalchemy', source_script=__file__)


# -------------------------------------------------------------------------------------------------------------------
#                                               DATABASE SCHEMA
#
def scripts_chat():
  return [sqlalchemy.Column('script_chat_id', sqlalchemy.Integer, primary_key=True, autoincrement=True),
          sqlalchemy.Column('report_name', sqlalchemy.String, nullable=False),
          sqlalchemy.Column('script_name', sqlalchemy.String, nullable=False),
          sqlalchemy.Column('message', sqlalchemy.Text, nullable=False),
          sqlalchemy.Column('usr_name', sqlalchemy.String, nullable=False),
          sqlalchemy.Column('html_code', sqlalchemy.String, nullable=False),
          sqlalchemy.Column('lst_mod_dt', sqlalchemy.DateTime, nullable=False, default=datetime.datetime.utcnow())]


def system_comments():
  return [sqlalchemy.Column('com_id', sqlalchemy.Integer, primary_key=True),
    sqlalchemy.Column('environment', sqlalchemy.String, nullable=False),
    sqlalchemy.Column('report', sqlalchemy.String, nullable=False),
    sqlalchemy.Column('htmlCode', sqlalchemy.String, nullable=False),
    sqlalchemy.Column('username', sqlalchemy.String, nullable=False),
    sqlalchemy.Column('visibility', sqlalchemy.String, nullable=False),
    sqlalchemy.Column('comment', sqlalchemy.String, nullable=False),
    sqlalchemy.Column('pmts', sqlalchemy.String),
    sqlalchemy.Column('lst_mod_dt', sqlalchemy.DateTime, default=datetime.datetime.utcnow(), nullable=True)]


def system_comments_replies():
  return [sqlalchemy.Column('reply_id', sqlalchemy.Integer, autoincrement=True, primary_key=True),
    sqlalchemy.Column('master_com_id', sqlalchemy.String),
    sqlalchemy.Column('environment', sqlalchemy.String, nullable=False),
    sqlalchemy.Column('report', sqlalchemy.String, nullable=False),
    sqlalchemy.Column('htmlCode', sqlalchemy.String, nullable=False),
    sqlalchemy.Column('username', sqlalchemy.String, nullable=False),
    sqlalchemy.Column('visibility', sqlalchemy.String, nullable=False),
    sqlalchemy.Column('comment', sqlalchemy.String, nullable=False),
    sqlalchemy.Column('pmts', sqlalchemy.String),
    sqlalchemy.Column('lst_mod_dt', sqlalchemy.DateTime, default=datetime.datetime.utcnow(), nullable=True)]


def system_user_comments():
  return [sqlalchemy.Column('auth_com_id', sqlalchemy.Integer, autoincrement=True, primary_key=True),
    sqlalchemy.Column('com_id', sqlalchemy.Integer, nullable=False),
    sqlalchemy.Column('auth_name', sqlalchemy.String, nullable=False),
    sqlalchemy.Column('lst_mod_dt', sqlalchemy.DateTime, default=datetime.datetime.utcnow(), nullable=True)]


class Comments(Html.Html):
  name, category, callFnc = 'Comment', 'Text', 'comments'
  __reqCss, __reqJs = ['jquery-scrollbar'], ['jquery-scrollbar']
  # _grpCls = CssGrpClsText.CssClassComment

  def __init__(self, report, htmlCode, recordset, title, pmts, db_service, width, height, httpCodes, readonly, profile):
    if httpCodes is None:
      self.httpCodeStr = ''
    else:
      self.httpCodeStr = '#%s' % '#'.join([report.http[code] for code in httpCodes])
    com_table = 'system_comments'
    reply_table = 'system_comments_replies'
    self.privacy = 'public' if not db_service else db_service.get('privacy', 'public')
    self.code = htmlCode
    super(Comments, self).__init__(report, recordset, width=width[0], widthUnit=width[1], code=htmlCode, profile=profile)
    self.css({"clear": 'both', "display": "inline-block", 'margin': '5px 0', 'padding': '0'})
    self.pmts, self._height, self.readonly = pmts, "%s%s" % (height[0], height[1]), readonly
    self.add_title(title, options={'content_table': False})
    if db_service is not None:
      if 'db' not in db_service:
        raise Exception('The db keyword needs to be defined with a proper epyk db Object (report.db.public)')

      com_table = db_service.get('com_table', 'system_comments')
      reply_table = db_service.get('reply_table', 'system_comments_replies')
      db_service['db'].table_create(com_table, system_comments())
      db_service['db'].table_create(reply_table, system_comments_replies())
      db_service['db'].table_create(db_service.get('user_coms', 'system_user_comments'), system_user_comments())
      db = db_service['db']
    else:
      db_service = {}
      db = self._report.data.db.public
      db.table_create(com_table, system_comments())
      db.table_create(reply_table, system_comments_replies())
      db.table_create(db_service.get('user_coms', 'system_user_comments'), system_user_comments())
    # Bind the database services associated to this component
    db_service.setdefault('service', '%s/comment/add' % self._report._urlsApp['report'])
    db_service.setdefault('reply_service', '%s/comment/reply' % self._report._urlsApp['report'])
    self.com_table = com_table
    self.reply_table = reply_table
    self.db_service = db_service
    for rec in db.select([com_table]).where([db.column(com_table, 'environment') == self._report.run.report_name,
                                             db.column(com_table, 'report') == self._report.run.script_name,
                                             db.column(com_table, 'htmlCode') == '%s%s' % (self.code, self.httpCodeStr)]).order_by(db.asc(db.column(com_table, 'lst_mod_dt'))).fetch():
      com_replies = []
      for reply in db.select([reply_table]).where([db.column(reply_table, 'environment') == self._report.run.report_name,
                                             db.column(reply_table, 'report') == self._report.run.script_name,
                                             db.column(reply_table, 'master_com_id') == rec['com_id'],
                                             db.column(reply_table, 'htmlCode') == '%s%s' % (self.code, self.httpCodeStr)]).order_by(db.asc(db.column(reply_table, 'lst_mod_dt'))).fetch():
        com_replies.append(dict(reply))
      recordset.append({'id': rec['com_id'], 'value': rec['comment'], 'user': rec['username'], 'time': datetime.date.strftime(rec['lst_mod_dt'], '%Y-%m-%d %H:%M:%S'), 'replies': com_replies, 'nb_replies': len(com_replies)})

  @property
  def jqId(self): return "$('#%s_comms')" % self.htmlId

  @property
  def val(self): return "$(this).val()"

  def onDocumentLoadFnc(self):
    self.addGlobalFnc("%s(htmlObj, data, jsStyles)" % self.__class__.__name__, '''htmlObj.empty();
      data.forEach(function(rec){addComment(htmlObj, rec.value, rec.user, rec.time, rec.replies, rec.nb_replies, rec.id)});
      $('.scroll_content').mCustomScrollbar()''', 'Javascript Object builder')

  def save(self):
    if self._report.run.current_user != 'local':
      data = ''' 'env': "%s" , 'report': "%s", 'htmlCode': "%s", "table_name" : "%s",
              'username': "%s", 'visibility': "%s", 'val': %s, 'path': JSON.stringify('%s'), 'httpCodes': "%s", 'timestamp': timeStamp ''' % (self._report.run.report_name, self._report.run.script_name,
                                                                              self.code, self.com_table, self._report.run.current_user, self.privacy, self.val,
                                                                              self._report.run.local_path, self.httpCodeStr)
      return self._report.jsPost(self.db_service['service'], data, isPyData=False)

    return ''

  def saveReply(self):
    if self._report.run.current_user != 'local':
      data = ''' 'env': "%s" , 'report': "%s", 'htmlCode': "%s", "table_name" : "%s",
              'username': "%s", 'visibility': "%s", 'val': %s, 'path': JSON.stringify('%s'), 'httpCodes': "%s", 'master_com_id': parentId ''' % (self._report.run.report_name, self._report.run.script_name,
                                                                              self.code, self.reply_table, self._report.run.current_user, self.privacy, self.val,
                                                                              self._report.run.local_path, self.httpCodeStr)
      return self._report.jsPost(self.db_service['reply_service'], data, isPyData=False)

    return ''

  def jsAdd(self, jsData='data', jsDataKey=None, isPyData=False, pmts=None, jsParse=False):
    jsData = self._jsData(jsData, jsDataKey, jsParse, isPyData)
    return "addComment(%s, %s, '%s', Today())" % (self.jqId, jsData, self._report.user)

  def __str__(self):
    self._report.jsOnLoadFnc.add("$('#%(htmlId)s_text').on('keypress', function (e) {if(e.which === 13){var timeStamp = new Date().getTime(); %(save)s; addComment($('#%(htmlId)s_comms'), $(this).val(), '%(user)s', Today(), [], 0, '%(user)s_' + timeStamp); $(this).val('')}})" % {'save': self.save(), 'htmlId': self.htmlId, 'user': self._report.run.current_user})
    self.addGlobalFnc('replyComment(htmlObj, user, parentId)', '''$('.comments_reply_inputs').remove(); 
    var inputReply = $('<input class="comments_reply_inputs" style="display:block;margin-left:20px" id=' + htmlObj.attr("id") + '_text type="text" placeholder="Add Reply" />');
    htmlObj.append(inputReply);
    inputReply.on('keypress', function (e) {if(e.which === 13){%(saveReply)s; addReply(htmlObj, $('#%(htmlId)s_comms'), $(this).val(), '%(user)s', Today()); $(this).remove()}})''' % {'saveReply': self.saveReply(), 'htmlId': self.htmlId, 'user': self._report.run.current_user})
    self.addGlobalFnc('addComment(htmlObj, data, user, timeStamp, replies, nbReplies, comId)', r'''
      var countComs = parseInt($('#'+ htmlObj.attr('id') + '_count').text()) + 1 + nbReplies; 
      var div = $("<div style='margin:5px;padding:0;text-align:left'>");
      div.append("<p style='margin:0;padding:0;display:inline-block;color:%(color)s;font-weight:bold'>"+ user +"</p>");
      div.append("<p style='margin:0;padding:0;display:inline-block;margin-left:10px;color:%(grey)s'>"+ timeStamp +"</p>");
      var replyCount = $("<p style='margin:0;padding:0;display:inline-block;margin-left:10px;color:%(color)s;font-weight:bold'> Replies: </p><p id=" + comId + "_nbRep style='display:inline-block;margin-left:5px;color:%(color)s;font-weight:bold'> "+ 0 +"</p>");
      div.append(replyCount);
      var button = $("<button id=" + comId + "_button style='margin:0;padding:0;display:inline-block;margin-left:10px;color:#3bb194;background:none;border:none'> Reply </button>");
      div.append(button);
      var commentTxt = $("<p id=" + comId +" style='margin:0 5px 20px 5px;padding:0;text-align:left'>"+ data +"</p>");
      htmlObj.prepend(commentTxt);
      button.on('click', function(e) {replyComment(commentTxt, user, comId);});
      htmlObj.prepend(div); 
      replies.forEach(function(rec) {addReply(commentTxt, $('#%(htmlId)s_comms'), rec.comment, rec.username, rec.lst_mod_dt)});
      $('#'+ htmlObj.attr('id') +'_count').text(countComs)''' % {'htmlId': self.htmlId, r'color': self._report.theme.greys[-1], 'grey': self._report.theme.greys[6]})
    self.addGlobalFnc('addReply(htmlObj, parentObj, data, user, timeStamp)', '''
      var replyCount = parseInt($('#'+ $.escapeSelector(htmlObj.attr('id')) + '_nbRep').text()) + 1;
      if (replyCount == 1) 
      {
        htmlObj.append("<div id=" + htmlObj.attr('id') + "_replies></div>");
      }
      var div = $("<div style='margin-left:20px;padding:0;text-align:left'>");
      div.append("<p style='margin:0;padding:0;display:inline-block;color:%(color)s;font-weight:bold'>"+ user +"</p>");
      div.append("<p style='margin:0;padding:0;display:inline-block;margin-left:10px;color:%(grey)s'>Replied At:</p>");
      div.append("<p style='margin:0;padding:0;display:inline-block;margin-left:10px;color:%(grey)s'>"+ timeStamp +"</p>");
      $("#" + $.escapeSelector(htmlObj.attr('id')) + "_replies").prepend("<p style='margin:0 5px 20px 20px;padding:0;text-align:left'>"+ data +"</p>");
      $("#" + $.escapeSelector(htmlObj.attr('id')) + "_replies").prepend(div);
      var countTotalComs = parseInt($('#'+ parentObj.attr('id') + '_count').text()) + 1;
      $('#'+ parentObj.attr('id') +'_count').text(countTotalComs);
      $('#'+ $.escapeSelector(htmlObj.attr('id')) + '_nbRep').text(replyCount);
    ''' % {'color': self._report.theme.greys[-1], 'grey': self._report.theme.greys[6]})
    inputTag = ''
    if not self.readonly:
      inputTag = '<input spellcheck=”false” style="display:block" id="%s_text" type="text" placeholder="Add public comment" />' % self.htmlId
    return '''
      <div %(attr)s>
        <span><p style="display:inline-block;margin:0;padding:0;cursor:pointer" id="%(htmlId)s_comms_count">0</p> Comments <i style="margin:0 5px 0 20px;cursor:pointer;display:inline-block" class="fas fa-sort-amount-up"></i>Sort by</span>
        %(inputTag)s
        <div class='scroll_content' style="margin:0;padding:5px 0;height:%(height)s;overflow:auto">
          <div id="%(htmlId)s_comms"></div>
        </div>
      </div>
      ''' % {'attr': self.get_attrs(pyClassNames=self.defined), 'htmlId': self.htmlId, 'height': self._height,
             'inputTag': inputTag}

  # -----------------------------------------------------------------------------------------
  #                                    EXPORT OPTIONS
  # -----------------------------------------------------------------------------------------
  def to_word(self, document): pass
  def to_xls(self, workbook, worksheet, cursor): pass
  def to_ppt(self, workbook, worksheet, cursor): pass


class Chat(Html.Html):
  name, category, callFnc = 'Chat', 'Messaging', 'chat'
  __reqCss, __reqJs = ['jquery-scrollbar'], ['jquery-scrollbar']
  # _grpCls = CssGrpClsText.CssClassComment

  def __init__(self, report, htmlCode, title, pmts, dbService, width, height, httpCodes, readonly, profile, chatOptions):
    self.readonly, self._height = readonly, "%s%s" % (height[0], height[1])
    if chatOptions is None:
      chatOptions = {}
    self.chatOptions = chatOptions
    if httpCodes is None:
      self.httpCodeStr = ''
    else:
      self.httpCodeStr = '#%s' % '#'.join([report.http[code] for code in httpCodes])
    # Get the database service
    db = report.data.db.public
    db.table_create("scripts_chat", scripts_chat())
    recordSet = []
    for rec in db.select("scripts_chat").where([db.column("scripts_chat", 'lst_mod_dt') > datetime.datetime.utcnow().date()]).records:
      rec["time"] = str(rec["lst_mod_dt"]).split(".")[0]
      recordSet.append(rec)
    self.privacy = 'public' if not dbService else dbService.get('privacy', 'public')
    super(Chat, self).__init__(report, recordSet, width=width[0], widthUnit=width[1], code=htmlCode, profile=profile)
    # Add internal HTML components
    self.add_title(title, options={'content_table': False})
    self._report.jsImports.add('socket.io')
    self._report.jsOnLoadFnc.add("var socket = io.connect('%s')" % self._report.run.url_root)
    self._report.jsOnLoadFnc.add("socket.on('message_%s_%s_%s', function(data) {%s})" % (
      self._report.run.report_name, self._report.run.script_name, self.htmlId, self.jsAppend()))

  @property
  def jqId(self): return "$('#%s_comms')" % self.htmlId

  @property
  def val(self): return "$(this).val()"

  def onDocumentLoadFnc(self):
    self.addGlobalFnc("%s(htmlObj, data, jsStyles)" % self.__class__.__name__, '''htmlObj.empty();
      data.forEach(function(rec){addChatMessage(htmlObj, rec.message, rec.usr_name, rec.time, true)});
      htmlObj.prepend("<hr style='background:%(color)s;margin:10px 10px 0 10px'/>");
      $('.scroll_content').mCustomScrollbar()''' % {'color': self._report.theme.greys[5]}, 'Javascript Object builder')

  def jsAppend(self, jsData='data', jsDataKey=None, isPyData=False, pmts=None, jsParse=False):
    jsData = self._jsData(jsData, jsDataKey, jsParse, isPyData)
    return "addChatMessage(%(jqId)s, %(data)s.text, %(data)s.user, Today(), false)" % {"jqId": self.jqId, "data": jsData}

  def __str__(self):
    self._report.jsOnLoadFnc.add('''
      $('#%(htmlId)s_text').on('keypress', function (e) {if(e.which === 13){
        var timeStamp = new Date().getTime(); var content = $(this).val();
        var data = {user: '%(user)s', text: content, timestamp: timeStamp, ext_system: %(chatOptions)s};
        %(postMessage)s; $(this).val('')}})
      ''' % {'htmlId': self.htmlId, 'user': self._report.run.current_user, 'chatOptions': json.dumps(self.chatOptions),
             "postMessage": self._report.jsPost('%s/socket/%s/%s/%s' % (self._report._urlsApp['index'].replace("/index", ""), self.htmlId, self._report.run.report_name, self._report.run.script_name))})
    self.addGlobalFnc('addChatMessage(htmlObj, data, user, timeStamp, oldMsg)', r'''
      var div = $("<div style='margin:5px;padding:0;text-align:left'>");
      var color = "%(color)s"; if(oldMsg){color = "%(grey)s"};
      div.append("<p style='margin:0;padding:0;display:inline-block;color:"+ color +";font-weight:bold'>"+ user +"</p>");
      div.append("<p style='margin:0;padding:0;display:inline-block;margin-left:10px;color:%(grey)s'>"+ timeStamp +"</p>");
      var commentTxt = $("<p style='margin:0 5px 20px 5px;padding:0;color:"+ color +";text-align:left'>"+ data +"</p>");
      htmlObj.prepend(commentTxt); htmlObj.prepend(div)''' % {'htmlId': self.htmlId, 'color': self._report.theme.greys[-1], 'grey': self._report.theme.greys[6]})
    inputTag = ''
    if not self.readonly:
      inputTag = '<input spellcheck="false" style="display:block" id="%s_text" type="text" placeholder="Write a message" />' % self.htmlId
    return '''
      <div %(attr)s>
        <span><p style="display:inline-block;margin:0;padding:0;cursor:pointer" id="%(htmlId)s_comms_count">0</p> Comments <i style="margin:0 5px 0 20px;cursor:pointer;display:inline-block" class="fas fa-sort-amount-up"></i>Sort by</span>
        %(inputTag)s
        <div class='scroll_content' style="margin:0;padding:5px 0;height:%(height)s;overflow:auto">
          <div id="%(htmlId)s_comms"></div>
        </div>
      </div>
      ''' % {'attr': self.get_attrs(pyClassNames=self.defined), 'htmlId': self.htmlId, 'height': self._height,
             'inputTag': inputTag}

  # -----------------------------------------------------------------------------------------
  #                                    EXPORT OPTIONS
  # -----------------------------------------------------------------------------------------
  def to_word(self, document): pass
  def to_xls(self, workbook, worksheet, cursor): pass
  def to_ppt(self, workbook, worksheet, cursor): pass


class Bot(Html.Html):
  """

  Three main states for the bot:
    - Message
    - Question
    - Advice / Information

  """
  name, category, callFnc = 'Bot', 'Messaging', 'bot'

  def __init__(self, report, htmlCode, name, pmts, dbService, width, height, httpCodes, profile, botOptions):
    super(Bot, self).__init__(report, [], width=width[0], widthUnit=width[1], code=htmlCode, profile=profile)
    self.css({"position": 'fixed', "bottom": 0, 'margin': '10px', 'background': self._report.theme.greys[0],
              "right": 0, "display": "block", "padding": "5px", 'border': "1px solid %s" % self._report.theme.success[1],
              "z-index": 200})
    self.name, self.height = name, height

  def __str__(self):
    self.addGlobalFnc('addBotMessage(htmlObj)', '''
      htmlObj.empty();
      var text = jQuery('<textarea type=\\"text\\" style=\\"width:100%%;height:%(height)spx\\" spellcheck=\\"false\\"></textarea>');
      htmlObj.append(text)''' % {"height": self.height[0] - 55})
    self.addGlobalFnc('addBotAdvices(htmlObj, message)', '''
      htmlObj.empty();
      var text = jQuery('<div>'+ message +'</div>');
      htmlObj.append(text)''' % {"height": self.height[0] - 55})
    self.addGlobalFnc('addBotQuestion(htmlObj, message, answers)', '''
      htmlObj.empty();
      var text = jQuery('<div>'+ message +'</div>');
      htmlObj.append(text)''' % {"height": self.height[0] - 55})
    # onclick="addBotMessage(jQuery('#content_bot'))"
    return '''
       <div %(attr)s>
          <p style="font-size:18px;width:100%%;margin:0;display:block">
            <i class="far fa-comments" style="font-size:12px;float:left;margin-right:5px"></i>
            <span onclick="addBotAdvices(jQuery('#content_bot'), 'text')" style="font-size:12px;cursor:pointer;float:left">%(name)s</span>
            <i onclick="jQuery('#content_bot').empty()" class="far fa-minus-square" style="font-size:12px;float:right;cursor:pointer"></i>
          </p>
          <div id="content_bot" style="clear:both;display:block"></div>
       </div>   
    ''' % {'attr': self.get_attrs(), "name": self.name, "height": self.height[0] - 55}

  # -----------------------------------------------------------------------------------------
  #                                    EXPORT OPTIONS
  # -----------------------------------------------------------------------------------------
  def to_word(self, document): pass
  def to_xls(self, workbook, worksheet, cursor): pass
  def to_ppt(self, workbook, worksheet, cursor): pass


class Alert(Html.Html):
  __reqJs, __reqCss = ['bootstrap'], ['bootstrap']
  name, category, callFnc = 'Alert', 'Messagings', 'alert'
  marginTop = 90

  def __init__(self, report, title, value, category, width, height, close_button=False,
               background_color=None, color='black', htmlCode=None, dataSrc=None, profile=False):
    super(Alert, self).__init__(report, {'title': title, 'value': value, 'closure': close_button}, width=width[0],
                                      widthUnit=width[1], height=height[0], heightUnit=height[1], code=htmlCode,
                                      dataSrc=dataSrc, profile=profile)
    self.close_button = close_button
    self.attr["class"].add('alert')
    self.attr["class"].add('alert-%s' % category.lower())
    heightSpace = 70 if self.close_button else 10
    self._jsStyles = {'class': ['alert', 'alert-%s' % category.lower()], "css": {'position': 'fixed', 'z-index': '290', 'right': '5px',
                      'width': '%s%s' % (width[0], width[1]), 'height': "%s%s" % (height[0], height[1]),
                      'color': color,
                      'top': '%spx' % (self.marginTop + (report._props["ui"]["notifications"]["count"] - 1) * heightSpace)}}
    if background_color is not None:
      self._jsStyles['background'] = background_color
    self.css(self._jsStyles["css"])
    if self.close_button:
      self._jsStyles['class'] += ['alert-dismissable']
      self.style.addCls('alert-dismissable')

  def onDocumentLoadFnc(self):
    self.addGlobalFnc("%s(htmlObj, data, jsStyles)" % self.__class__.__name__, ''' 
      if (htmlObj.length == 0){htmlObj = $("<div class='"+ jsStyles.class.join(" ") +"'></div>"); htmlObj.css(jsStyles.css); $('body').append(htmlObj)}
      htmlObj.empty();
      if((data.closure) || (data.closure == undefined)){htmlObj.append('<a href="#" class="close" data-dismiss="alert" aria-label="close">&times;</a>')}
      htmlObj.append('<strong>'+ data.title +'!</strong>&nbsp;'+ data.value)''')

  def __str__(self):
    return "<div %s></div>" % self.get_attrs(pyClassNames=self.pyStyle)


class News(Html.Html):
  name, category, callFnc = 'News', 'Messaging', 'news'
  __reqJs, __reqCss = ["jqueryui"], ["jqueryui"]

  def __init__(self, report, title, value, label, link_script, icon, width, height, htmlCode, profile):
    super(News, self).__init__(report, value, width=width[0], widthUnit=width[1],
                               height=height[0], heightUnit=height[1], code=htmlCode, profile=profile)
    self.css({"padding": '5px', "display": 'none', 'position': 'fixed', 'border': '1px solid %s' % self._report.theme.success[1],
              "background": self._report.theme.greys[0], 'bottom': '20px', 'right': '20px'})
    # Add internal HTML component to the new feed
    self.add_label(label)
    self.add_title(title, options={'content_table': False})
    self.add_link(link_script, position="after")

  def onDocumentLoadFnc(self):
    self.addGlobalFnc("%s(htmlObj, data, jsStyles)" % self.__class__.__name__, "htmlObj.find('p#content').html(data)")

  def jsGenerate(self, jsData='data', jsDataKey=None, isPyData=False, jsParse=False, jsStyles=None, jsFnc=None):
    js_data = super(News, self).jsGenerate(jsData, jsDataKey, isPyData, jsParse, jsStyles, jsFnc)
    return "%(js_data)s;%(jqId)s.show().delay(1000);%(jqId)s.fadeOut(1000)" % {"js_data": js_data, "jqId": self.jqId}

  def __str__(self):
    return "<div %s><p id='content'></p></div>" % self.get_attrs(pyClassNames=self.pyStyle)



"""
Module in charge of the PivotTable library

Related Pages:

		https://pivottable.js.org/examples/
"""

import json

from epyk.core.html import Html
from epyk.core.js.objects import JsPivotFncs

# The list of CSS classes
# from epyk.core.css.styles import CssGrpClsTable

extensions = {
  'sub-total': {'jsImports': ['pivot-sub-total']},
  'c3': {'jsImports': ['pivot-c3']},
}


class PivotTable(Html.Html):
  __reqJs, __reqCss = ["pivot"], ["pivot"]
  name, category, callFnc = 'Pivot Table', 'Tables', 'pivot'
  # _grpCls = CssGrpClsTable.CssStylesPivot

  def __init__(self, report, recordSet, rows, cols, valCol, title, tableOptions, width, height, aggOptions, rendererName,
               htmlCode, dataSrc, helper, profile):
    super(PivotTable, self).__init__(report, [], width=width[0], widthUnit=width[1], height=height[0], heightUnit=height[1],
                                     code=htmlCode, dataSrc=dataSrc, profile=profile)
    # Add the extra HTML components
    self.add_title(title, options={'content_table': False})
    self.add_helper(helper)
    # to add all the columns in the table if nothing defined
    self.aggOptions, self.dsc, self.tableOptions, self.addinOptions = aggOptions, '', dict(tableOptions), []
    self.__pivot = {'cols': [] if cols is None else cols, 'rows': [] if rows is None else rows, 'vals': [] if valCol is None else valCol,
                    'aggregatorName': aggOptions['name'], 'rendererName': rendererName}
    self.__aggFncs = dict(JsPivotFncs.getAggFnc())
    self.data = recordSet
    self.data.attach(self)
    self.css({"overflow": 'auto', 'margin': '0 auto'})
    self.addGlobalFnc("numberWithCommas(x)", 'return x.toString().replace(/\B(?=(\d{3})+(?!\d))/g, ",")')
    if not self.tableOptions.get("editable", True):
      self.addinOptions.append("window['options_%s'].showUI = false" % self.htmlId)
      del self.tableOptions["editable"]

    if self.tableOptions.get('sub-total'):
      self._report.jsImports.add('pivot-sub-total')
      self.addinOptions.append("window['options_%s'].dataClass = $.pivotUtilities.SubtotalPivotData" % self.htmlId)
      self.addinOptions.append("window['options_%s'].renderers = $.pivotUtilities.subtotal_renderers" % self.htmlId)
      self.addinOptions.append("window['options_%s'].rendererName = 'Table With Subtotal'" % self.htmlId)
      del tableOptions['sub-total']

    # https://pivottable.js.org/examples/onrefresh.html
    for k, v in self.tableOptions.items():
      if k == 'sorters':
        row = []
        for i, j in v.items():
          if isinstance(j, list):
            row.append("'%s': $.pivotUtilities.sortAs(%s)" % (i, j))
          else:
            row.append("'%s': %s" % (i, j))
        self.addinOptions.append("window['options_%(htmlId)s'].%(opt)s = {%(val)s}" % {'htmlId': self.htmlId, 'opt': k, 'val': ",".join(row)})
      else:
        self.addinOptions.append("window['options_%(htmlId)s'].%(opt)s = %(val)s" % {'htmlId': self.htmlId, 'opt': k, 'val': json.dumps(v)})

  @property
  def jqId(self):
    return "$('#%s div')" % self.htmlId

  def addAggregator(self, aggCls):
    """
    Add on the fly new aggregation logic to the pivot table. This will allow the creation of bespoke aggregation functions.
    Those functions will be structure in the way they can be added to the core framework easily.
    Aggregators should follow some rules and they should defined some mandatory parameters (name, keyAgg, push, value, format).

    Example
      class JsPivotSumAgg(object):
        name = "New Sum Agg"
        keyAgg, key2Agg = 0, None
        push = 'this.keyAgg += parseFloat(record[attributeArray[0]]) * 2'
        value = 'return this.keyAgg'
        format = 'return numberWithCommas(x.toFixed(%(digits)s))'

    Related Pages:

			https://github.com/nicolaskruchten/pivottable/wiki/Aggregators
    """
    if not ':dsc:' in aggCls.__doc__:
      raise Exception("The new Aggregator class %s must have a doc string with a :dsc: field !" % aggCls.name)

    if aggCls.name in self.__aggFncs:
      raise Exception("Duplicated Name - Aggregator %s cannot be replaced !!!" % aggCls.name)

    self.__aggFncs[aggCls.name] = type(aggCls.__name__, (aggCls, JsPivotFncs.JsPivotAggFnc), {})()
    return self

  def jsGenerate(self, jsData='data', jsDataKey=None, isPyData=False, jsId=None):
    return """%(jqId)s.pivotUI(%(jsData)s, window['options_%(htmlId)s'])""" % {'jqId': self.jqId, 'jsData': self.data.setId(jsData).getJs(), 'htmlId': self.htmlId}

  def onDocumentReady(self):
    jsAggFncs = "{%s}" % ", ".join(["'%s': function(attributeArray) {return function(data, rowKey, colKey) {return %s}}" % (name, aggFncs.toJs(self.aggOptions)) for name, aggFncs in self.__aggFncs.items()])
    profile = self.profile if self.profile is not None else getattr(self._report, 'PROFILE', False)
    preFnc, endFnc = "", ""
    if profile:
      preFnc, endFnc = "var t0 = performance.now()", "console.log('|%s|%s|'+ (performance.now()-t0))" % (self.__class__.__name__, self.htmlId)
    self._report.jsOnLoadFnc.add(""" %(preFnc)s;
      var tpl = $.pivotUtilities.aggregatorTemplates; window['options_%(htmlId)s'] = %(options)s; %(addinOptions)s;
      window['options_%(htmlId)s'].aggregators = %(agg)s;
      window['options_%(htmlId)s'].onRefresh = function (config) {
          %(jqId)s.find('.pvtVal').each(function( index, items ) {
            if (parseFloat(items.innerText.replace(',', '.')) < 0){ $(items).css('color', 'red')}});
          %(jqId)s.find('.pvtTotal').each(function( index, items ) {
            if (parseFloat(items.innerText.replace(',', '.')) < 0){ $(items).css('color', 'red')}});
          %(jqId)s.find('.pvtGrandTotal').each(function( index, items ) {
            if (parseFloat(items.innerText.replace(',', '.')) < 0){ $(items).css('color', 'red')}});
      } """ % {'jqId': self.jqId, 'options': json.dumps(self.__pivot), 'agg': jsAggFncs, 'htmlId': self.htmlId,
               'addinOptions': ";".join(self.addinOptions), "preFnc": preFnc})
    self._report.jsOnLoadFnc.add("%s;%s" % (self.jsGenerate(None), endFnc))

  def onDocumentLoadFnc(self): return True

  def setTableCss(self, cssClss):
    """
    Add a style to the datatable. This can be used to change some specific part of the table (for example the header)

    Example
      class CssTableHeader(object):
        __style = [{'attr': 'background', 'value': 'grey'}]
        childrenTag = 'th'

      tb.setTableCss(CssTableHeader)

    :return: The Python Datatable object
    """
    import inspect

    if not isinstance(cssClss, list):
      cssClss = [cssClss]
    for cssCls in cssClss:
      if inspect.isclass(cssCls):
        self._report.cssObj.addPy(cssCls)
        cssCls = cssCls.__name__
      clssMod = self._report.style.add(cssCls)
      if clssMod is not None:
        self._report.style.cssCls(cssCls)
    return self

  def __str__(self):
    return '<div %(strAttr)s><div></div></div>%(helper)s' % {'strAttr': self.get_attrs(pyClassNames=self.defined), "helper": self.helper}


  # -----------------------------------------------------------------------------------------
  #                                    MARKDOWN SECTION
  # -----------------------------------------------------------------------------------------
  @classmethod
  def matchMarkDownBlock(cls, data):
    return True if data[0].strip().startswith("---Pivot") else None

  @staticmethod
  def matchEndBlock(data):
    return data.endswith("---")

  @classmethod
  def convertMarkDownBlock(cls, data, report=None):
    return []

  def jsMarkDown(self): return ""

